'''
See COPYRIGHT.md for copyright information.
'''
import os, io, sys, json
from collections import defaultdict, OrderedDict
from arelle import XbrlConst
from arelle.ModelObject import ModelObject
from arelle.ModelDtsObject import ModelResource, ModelRelationship
from arelle.ModelValue import QName
from arelle.Aspect import Aspect, aspectStr, aspectModelAspect
from arelle.ModelRenderingObject import (ResolutionException, DefnMdlTable, DefnMdlBreakdown,
                                         DefnMdlDefinitionNode, DefnMdlClosedDefinitionNode, DefnMdlRuleDefinitionNode,
                                         DefnMdlRelationshipNode, DefnMdlAspectNode, DefnMdlOpenDefinitionNode,
                                         DefnMdlConceptRelationshipNode, DefnMdlDimensionRelationshipNode,
                                         StrctMdlNode, StrctMdlTableSet, StrctMdlTable, StrctMdlBreakdown, StrctMdlStructuralNode,
                                         OPEN_ASPECT_ENTRY_SURROGATE, ROLLUP_SPECIFIES_MEMBER, ROLLUP_IMPLIES_DEFAULT_MEMBER,
                                         ROLLUP_FOR_CONCEPT_RELATIONSHIP_NODE, ROLLUP_FOR_DIMENSION_RELATIONSHIP_NODE,
                                         ROLLUP_FOR_CLOSED_DEFINITION_NODE, ROLLUP_FOR_OPEN_DEFINITION_NODE,
                                         ROLLUP_FOR_DEFINITION_NODE,
                                         UNREPORTED_ASPECT_SORT_VALUE,
                                         aspectStrctNodes)
from arelle.PrototypeInstanceObject import FactPrototype
from arelle.PythonUtil import flattenSequence, flattenToSet
from arelle.formula.XPathContext import XPathException, FunctionArgType
from numbers import Number
NoneType = type(None)
EMPTY_LIST = []
EMPTY_DICT = {}

TRACE_RESOLUTION = False
TRACE_TABLE_STRUCTURE = True

RENDER_UNITS_PER_CHAR = 16 # nominal screen units per char for wrapLength computation and adjustment

def prod(seq):
    p = 1
    for n in seq:
        p *= n
    return p

def resolveTableStructure(view, viewTblELR):
    if isinstance(viewTblELR, DefnMdlTable):
        # called with a defnMdlTable instead of an ELR

        # find an ELR for this table object
        defnMdlTable = viewTblELR
        strctMdlTableSet = StrctMdlTableSet(defnMdlTable)
        for rel in view.modelXbrl.relationshipSet((XbrlConst.tableBreakdown, XbrlConst.tableBreakdownMMDD)).fromModelObject(defnMdlTable):
            # find relationships in table's linkrole
            view.defnSubtreeRelSet = view.modelXbrl.relationshipSet((XbrlConst.tableBreakdownTree, XbrlConst.tableBreakdownTreeMMDD), rel.linkrole)
            return resolveTableAxesStructure(view, strctMdlTableSet,
                                             view.modelXbrl.relationshipSet((XbrlConst.tableBreakdown, XbrlConst.tableBreakdownMMDD), rel.linkrole))
        # no relationships from table found
        return None

    # called with an ELR or list of ELRs
    tblBrkdnRelSet = view.modelXbrl.relationshipSet((XbrlConst.tableBreakdown, XbrlConst.tableBreakdownMMDD), viewTblELR)
    view.defnSubtreeRelSet = view.modelXbrl.relationshipSet((XbrlConst.tableBreakdownTree, XbrlConst.tableBreakdownTreeMMDD, XbrlConst.tableDefinitionNodeSubtree, XbrlConst.tableDefinitionNodeSubtreeMMDD), viewTblELR)
    if tblBrkdnRelSet is None or len(tblBrkdnRelSet.modelRelationships) == 0:
        view.modelXbrl.modelManager.addToLog(_("no table relationships for {0}").format(viewTblELR))
        return None

    # table name
    modelRoleTypes = view.modelXbrl.roleTypes.get(viewTblELR)
    if modelRoleTypes is not None and len(modelRoleTypes) > 0:
        view.roledefinition = modelRoleTypes[0].definition
        if view.roledefinition is None or view.roledefinition == "":
            view.roledefinition = os.path.basename(viewTblELR)
    try:
        for defnMdlTable in tblBrkdnRelSet.rootConcepts:
            view.rendrCntx = defnMdlTable.renderingXPathContext
            strctMdlTableSet = StrctMdlTableSet(defnMdlTable)
            referencedVariables = defnMdlTable.variableRefs()
            for tblParamRel in view.modelXbrl.relationshipSet((XbrlConst.tableParameter, XbrlConst.tableParameterMMDD)).fromModelObject(defnMdlTable):
                referencedVariables.add(tblParamRel.variableQname) # https://gitlab.xbrl.org/rendering/table-linkbase/-/issues/32#note_20551
            # determine number of tables based on cartesian product of multi-valued parameters
            varSeqValues = OrderedDict(( # order tables by lexicographic key order
                (name, flattenSequence(view.rendrCntx.inScopeVars.get(name,())))
                for name in sorted(referencedVariables,reverse=True))) # want last variable to iterate first
            varSeqIndx = dict((name,0) for name in varSeqValues)
            for tblNum in range(prod((len(seqVal) or 1)
                                     for name, seqVal in varSeqValues.items())):
                strctMdlTable = StrctMdlTable(strctMdlTableSet, defnMdlTable)
                for name, seqVal in varSeqValues.items():
                    if len(seqVal) > 0:
                        thisSeqVal = seqVal[varSeqIndx[name]]
                        view.rendrCntx.inScopeVars[name] = [thisSeqVal]
                        strctMdlTable.tblParamValues[name] = thisSeqVal
                resolveTableAxesStructure(view, strctMdlTable, tblBrkdnRelSet)
                # iterate paramSeqIndex
                for name, seqVal in varSeqValues.items():
                    varSeqIndx[name] += 1
                    if varSeqIndx[name] < len(seqVal):
                        view.rendrCntx.inScopeVars[name] = seqVal[varSeqIndx[name]]
                        break
                    varSeqIndx[name] = 0 # reset this param and set next one
                    if len(seqVal) > 0:
                        view.rendrCntx.inScopeVars[name] = seqVal[varSeqIndx[name]]
            return strctMdlTableSet
    except ResolutionException as ex:
        #traceback.print_exc()
        view.modelXbrl.error(ex.code, ex.message, exc_info=True, **ex.kwargs);

    return None

def resolveTableAxesStructure(view, strctMdlTable, tblBrkdnRelSet):
    view.dataCols = 0
    view.dataRows = 0
    view.dataFirstCol = 0
    view.dataFirstRow = 0
    view.colHdrNonStdRoles = []
    view.colHdrDocRow = False
    view.colHdrCodeRow = False
    view.colHdrRows = 0
    view.rowHdrNonStdRoles = []
    view.rowHdrCols = 0
    view.rowHdrColWidth = [0,]
    view.rowNonAbstractHdrSpanMin = [0,]
    view.rowHdrDocCol = False
    view.rowHdrCodeCol = False
    view.zAxisBreakdowns = 0
    view.zmostOrdCntx = None
    view.defnMdlTable = defnMdlTable = strctMdlTable.defnMdlNode
    view.aspectEntryObjectId = 0

    # must be cartesian product of top level relationships
    view.tblBrkdnRels = tblBrkdnRels = tblBrkdnRelSet.fromModelObject(defnMdlTable)

    for tblBrkdnRel in tblBrkdnRels:
        defnMdlBreakdown = tblBrkdnRel.toModelObject
        strctMdlTable.defnMdlBreakdowns[tblBrkdnRel.axis].append(defnMdlBreakdown)

    facts = view.modelXbrl.factsInInstance
    if facts:
        facts = defnMdlTable.filteredFacts(view.rendrCntx, view.modelXbrl.factsInInstance) # apply table filters
    # do z's first to set variables needed by x and y axes expressions
    for axis in ("z", "y", "x"):
        axisBrkdnRels = [r for r in tblBrkdnRels if r.axis == axis]
        for tblBrkdnRel in axisBrkdnRels:
            defnMdlBreakdown = tblBrkdnRel.toModelObject
            strctMdlBreakdown = resolveDefinition(view, strctMdlTable, defnMdlBreakdown, 0, facts, 0, axisBrkdnRels, axis=axis)
            # perform any supplemental leveling of child str mdl nodes
            descendantDepths = [getDepth(d) for d in strctMdlBreakdown.strctMdlChildNodes]
            if min(descendantDepths) != max (descendantDepths):
                addDescendantRollups(strctMdlBreakdown)
            # add any defaulted dimensions needed on leaf nodes
            addDefaultedDimensionsToLeafNodes(strctMdlBreakdown)
            if axis == "x":
                view.dataCols += strctMdlBreakdown.leafNodeCount
            elif axis == "y":
                view.dataRows += strctMdlBreakdown.leafNodeCount
            elif axis == "z":
                view.zAxisBreakdowns = len(axisBrkdnRels)
        if not axisBrkdnRels: # no breakdown rels
            strctMdlBreakdown = resolveDefinition(view, strctMdlTable, None, 1, facts, 0, axisBrkdnRels, axis=axis)
            strctMdlBreakdown.setHasOpenNode()
            if axis == "x":
                view.dataCols += strctMdlBreakdown.leafNodeCount
                strctMdlBreakdown.setHasOpenNode()
            elif axis == "y":
                view.dataRows += strctMdlBreakdown.leafNodeCount
                strctMdlBreakdown.setHasOpenNode()
    # check for clashing and unresolved tagSelectors
    strctMdlAxisNodes = defaultdict(list)
    def compileAlignedStrctNodes(strctNode):
        childNodes = strctNode.strctMdlChildNodes
        if not childNodes: # this is a leaf node
            if not isinstance(strctNode, StrctMdlBreakdown) and not strctNode.defnMdlNode.isAbstract:
                strctMdlAxisNodes[strctNode.axis].append(strctNode)
        else:
            for childStrctNode in childNodes:
                compileAlignedStrctNodes(childStrctNode)
    for axis in ("z", "y", "x"):
        compileAlignedStrctNodes(strctMdlTable.strctMdlFirstAxisBreakdown(axis))
    for zStrctMdlNode in strctMdlAxisNodes.get("z", [None]):
        zAspectStrctNodes = aspectStrctNodes(zStrctMdlNode)
        for yStrctMdlNode in strctMdlAxisNodes.get("y", [None]):
            yAspectStrctNodes = aspectStrctNodes(yStrctMdlNode)
            for xStrctMdlNode in strctMdlAxisNodes.get("x", [None]):
                xAspectStrctNodes = aspectStrctNodes(xStrctMdlNode)
                stNodes = (zStrctMdlNode, yStrctMdlNode, xStrctMdlNode)
                cellTagSelectors = set(ts for sn in stNodes if sn for ts in sn.tagSelectors if ts)
                if {"table.periodStart","table.periodEnd"} & cellTagSelectors:
                    cellTagSelectors &= {"table.periodStart","table.periodEnd"}
                for _aspectStrctNodes in (zAspectStrctNodes.values(), yAspectStrctNodes.values(), xAspectStrctNodes.values()):
                    for aspectStrctNode in flattenToSet(_aspectStrctNodes):
                        strctNodeTags = aspectStrctNode.defnMdlNode.constraintSets.keys()
                        if len(cellTagSelectors & strctNodeTags) > 1:
                            view.modelXbrl.error("xbrlte:tagSelectorClash",
                                _("Cell has multiple tag selectors for same aspect node %(nodeId)s %(clashingSelectors)s"),
                                modelObject=stNodes, clashingSelectors=", ".join(cellTagSelectors & strctNodeTags), nodeId=aspectStrctNode.defnMdlNode.id)
                        for cellTagSelector in cellTagSelectors:
                            if cellTagSelector not in strctNodeTags and None not in strctNodeTags:
                                view.modelXbrl.error("xbrlte:noMatchingConstraintSet",
                                    _("Cell has missing constraint set for tag selector %(cellTagSelector)s, node %(nodeId)s"),
                                    modelObject=stNodes, cellTagSelector=cellTagSelector, nodeId=aspectStrctNode.defnMdlNode.id)

    view.colHdrTopRow = view.zAxisBreakdowns # need rest if combobox used (2 if view.zAxisRows else 1)
    for i in range(view.rowHdrCols):
        if view.rowNonAbstractHdrSpanMin[i]:
            lastRowMinWidth = view.rowNonAbstractHdrSpanMin[i] - sum(view.rowHdrColWidth[i] for j in range(i, view.rowHdrCols - 1))
            if lastRowMinWidth > view.rowHdrColWidth[view.rowHdrCols - 1]:
                view.rowHdrColWidth[view.rowHdrCols - 1] = lastRowMinWidth
    #view.rowHdrColWidth = (60,60,60,60,60,60,60,60,60,60,60,60,60,60)
    # use as wraplength for all row hdr name columns 200 + fixed indent and abstract mins (not incl last name col)
    view.rowHdrWrapLength = 200 + sum(view.rowHdrColWidth[:view.rowHdrCols + 1])
    if view.colHdrRows == 0:
        view.colHdrRows = 1 # always reserve a col header row even if no labels for col headers
    view.dataFirstRow = view.colHdrTopRow + view.colHdrRows + len(view.colHdrNonStdRoles)
    view.dataFirstCol = view.rowHdrCols + len(view.rowHdrNonStdRoles)
    #view.dataFirstRow = view.colHdrTopRow + view.colHdrRows + view.colHdrDocRow + view.colHdrCodeRow
    #view.dataFirstCol = 1 + view.rowHdrCols + view.rowHdrDocCol + view.rowHdrCodeCol
    #for i in range(view.dataFirstRow + view.dataRows):
    #    view.gridView.rowconfigure(i)
    #for i in range(view.dataFirstCol + view.dataCols):
    #    view.gridView.columnconfigure(i)
    # organize hdrNonStdRoles so code (if any) is after documentation (if any)
    for hdrNonStdRoles in (view.colHdrNonStdRoles, view.rowHdrNonStdRoles):
        iCodeRole = -1
        for i, hdrNonStdRole in enumerate(hdrNonStdRoles):
            if 'code' in os.path.basename(hdrNonStdRole).lower():
                iCodeRole = i
                break
        if iCodeRole >= 0 and len(hdrNonStdRoles) > 1 and iCodeRole < len(hdrNonStdRoles) - 1:
            del hdrNonStdRoles[iCodeRole]
            hdrNonStdRoles.append(hdrNonStdRole)
    if TRACE_RESOLUTION: print(
        f"dataCols {view.dataCols} dataRows {view.dataRows} dataFirstCol {view.dataFirstCol} dataFirstRow {view.dataFirstRow} "
        f"colHdrRows {view.colHdrRows} rowHdrCols {view.rowHdrCols} zAxisBreakdowns {view.zAxisBreakdowns} "
        f"colHdrTopRow {view.colHdrTopRow} colHdrRows {view.colHdrRows} colHdrNonStdRoles {len(view.colHdrNonStdRoles)} "
        f"rowHdrNonStdRoles {len(view.rowHdrNonStdRoles)}"
        )

def sortkey(obj):
    if isinstance(obj, ModelObject):
        return obj.objectIndex
    return obj

def childContainsOpenNodes(childStrctMdlNode):
    if childStrctMdlNode is None:
        return False
    if isinstance(childStrctMdlNode.defnMdlNode, DefnMdlAspectNode) \
       and (childStrctMdlNode.isLabeled \
            or any([node.isEntryPrototype(default=False) for node in childStrctMdlNode.strctMdlChildNodes])):
        # either the child structural node has a concrete header or it contains a structure
        # that has not yet a concrete value
        return True
    else:
        for node in childStrctMdlNode.strctMdlChildNodes:
            if childContainsOpenNodes(node):
                return True
    return False

def checkLabelWidth(view, strctMdlNode, subtreeRels, checkBoundFact=False):
    if strctMdlNode.axis == "y":
        # messages can't be evaluated, just use the text portion of format string
        label = strctMdlNode.header(lang=view.lang,
                                      returnGenLabel=not checkBoundFact,
                                      returnMsgFormatString=not checkBoundFact)
        if label:
            # need to et more exact word length in screen units
            widestWordLen = max(len(w) * RENDER_UNITS_PER_CHAR for w in label.split())
            # abstract only pertains to subtree of closed nodesbut not cartesian products or open nodes
            while strctMdlNode.depth >= len(view.rowHdrColWidth):
                view.rowHdrColWidth.append(RENDER_UNITS_PER_CHAR)
                view.rowNonAbstractHdrSpanMin.append(0)
            if strctMdlNode.defnMdlNode.isAbstract or not subtreeRels:
                if widestWordLen > view.rowHdrColWidth[strctMdlNode.depth]:
                    view.rowHdrColWidth[strctMdlNode.depth] = widestWordLen
            else:
                if widestWordLen > view.rowNonAbstractHdrSpanMin[strctMdlNode.depth]:
                    view.rowNonAbstractHdrSpanMin[strctMdlNode.depth] = widestWordLen

#def resolveDefinition(view, strctMdlNode, depth, facts, i=None, tblAxisRels=None, processOpenDefinitionNode=True, rollUpNode=None):
def resolveDefinition(view, strctMdlParent, defnMdlNode, depth, facts, iBrkdn=None, axisBrkdnRels=None, rollUpNode=None, axis=None, mergeAspects=None):
    if isinstance(defnMdlNode, (NoneType, DefnMdlBreakdown)):
        strctMdlNode = StrctMdlBreakdown(strctMdlParent, defnMdlNode, axis)
    else:
        if isinstance(defnMdlNode, (DefnMdlRelationshipNode,DefnMdlAspectNode)) or mergeAspects or defnMdlNode.isMerged:
            strctMdlNode = strctMdlParent # all children are added during relationship navigatio below
        else:
            strctMdlNode = StrctMdlStructuralNode(strctMdlParent, defnMdlNode)
    if strctMdlParent.axis:
        axis = strctMdlParent.axis

    subtreeRels = view.defnSubtreeRelSet.fromModelObject(defnMdlNode)

    if isinstance(defnMdlNode, (DefnMdlBreakdown, DefnMdlDefinitionNode)):
        try:
            try:
                ordCardinality, ordDepth = defnMdlNode.cardinalityAndDepth(strctMdlNode, handleXPathException=False)
            except (XPathException, FunctionArgType, ResolutionException) as ex:
                if (isinstance(defnMdlNode, DefnMdlRelationshipNode) and
                    type(ex) == FunctionArgType and ex.argNum == 5 and isinstance(ex.value, int) and ex.value >= 0):
                    view.modelXbrl.error("xbrlte:relationshipNodeTooManyGenerations",
                                         ex.expectedType,
                                         modelObject=(view.defnMdlTable,defnMdlNode), xlinkLabel=defnMdlNode.xlinkLabel, axis=defnMdlNode.localName)
                elif isinstance(defnMdlNode, DefnMdlRelationshipNode) and type(ex) == ResolutionException:
                    view.modelXbrl.error(ex.code,
                    _("Relationship node %(xlinkLabel)s expression not castable to required type (%(error)s)"),
                    modelObject=(view.defnMdlTable,defnMdlNode), xlinkLabel=defnMdlNode.xlinkLabel, axis=defnMdlNode.localName,
                    error=ex.message)
                else:
                    view.modelXbrl.error("xbrlte:expressionNotCastableToRequiredType",
                    _("Relationship node %(xlinkLabel)s expression not castable to required type (%(xpathError)s)"),
                    modelObject=(view.defnMdlTable,defnMdlNode), xlinkLabel=defnMdlNode.xlinkLabel, axis=defnMdlNode.localName,
                    xpathError=str(ex))
                return
            if (not defnMdlNode.isAbstract and
                isinstance(defnMdlNode, DefnMdlClosedDefinitionNode) and
                ordCardinality == 0 and not defnMdlNode.childrenCoverSameAspects):
                view.modelXbrl.error("xbrlte:closedDefinitionNodeZeroCardinality",
                    _("Closed definition node %(xlinkLabel)s does not contribute at least one structural node"),
                    modelObject=(view.defnMdlTable,defnMdlNode), xlinkLabel=defnMdlNode.xlinkLabel, axis=defnMdlNode.localName)
            nestedDepth = depth + ordDepth
            cartesianProductNestedArgs = [view, nestedDepth, axis, facts, axisBrkdnRels, iBrkdn]
            if axis == "x":
                if ordDepth:
                    if nestedDepth > view.colHdrRows:
                        view.colHdrRows = nestedDepth - 1
                hdrNonStdRoles = view.colHdrNonStdRoles
            elif axis == "y":
                if ordDepth:
                    if nestedDepth > view.rowHdrCols:
                        view.rowHdrCols = nestedDepth
                        if not isinstance(defnMdlNode, DefnMdlRelationshipNode):
                            # not quite sure of this condition, but Solvency (rule node) need one less col than a relationship node (german taxonomy)
                            view.rowHdrCols -= 1
                        for j in range(1 + ordDepth):
                            view.rowHdrColWidth.append(RENDER_UNITS_PER_CHAR)  # min width for 'tail' of nonAbstract coordinate
                            view.rowNonAbstractHdrSpanMin.append(0)
                    elif isinstance(strctMdlNode, StrctMdlBreakdown):
                        view.rowHdrCols = view.rowHdrCols + nestedDepth
                        for j in range(1 + ordDepth):
                            view.rowHdrColWidth.append(RENDER_UNITS_PER_CHAR)  # min width for 'tail' of nonAbstract coordinate
                            view.rowNonAbstractHdrSpanMin.append(0)
                    checkLabelWidth(view, strctMdlNode, subtreeRels, checkBoundFact=False)
                hdrNonStdRoles = view.rowHdrNonStdRoles
                # print(f"id {defnMdlNode.id} depth {depth} y axis ordDepth {ordDepth} nestedDepth {nestedDepth} rowHdrCols {view.rowHdrCols}")
            if axis in ("x", "y"):
                hdrNonStdPosition = -1  # where a match last occured
                for rel in view.modelXbrl.relationshipSet(XbrlConst.elementLabel).fromModelObject(defnMdlNode):
                    if isinstance(rel.toModelObject, ModelResource) and rel.toModelObject.role != XbrlConst.genStandardLabel:
                        labelLang = rel.toModelObject.xmlLang
                        labelRole = rel.toModelObject.role
                        if (labelLang == view.lang or labelLang.startswith(view.lang) or view.lang.startswith(labelLang)
                            or ("code" in labelRole)):
                            labelRole = rel.toModelObject.role
                            if labelRole in hdrNonStdRoles:
                                hdrNonStdPosition = hdrNonStdRoles.index(labelRole)
                            else:
                                hdrNonStdRoles.insert(hdrNonStdPosition + 1, labelRole)

            isCartesianProductExpanded = False
            if not isinstance(defnMdlNode, DefnMdlAspectNode):
                isCartesianProductExpanded = True
                # note: reduced set of facts should always be passed to subsequent open nodes
                childStrctNode = None
                hasUnreportedChildStrctNode = False
                for subtreeRel in subtreeRels:
                    childDefnMdlNode = subtreeRel.toModelObject

                    if childDefnMdlNode.isMerged:
                        '''
                        childSubtreeRels = view.defnSubtreeRelSet.fromModelObject(childDefnMdlNode)
                        for childSubtreeRel in childSubtreeRels:
                            mergedChildDefnMdlNode = childSubtreeRel.toModelObject
                            childStrctNode = resolveDefinition(view, strctMdlNode, mergedChildDefnMdlNode, depth+ordDepth, facts, iBrkdn, axisBrkdnRels, parentMerges=True)
                            for gStrctNode in strctMdlNode.strctMdlChildNodes:
                                for mergedAspect in childDefnMdlNode.aspectsCovered():
                                    aspect = childDefnMdlNode.aspectValue(view.rendrCntx, mergedAspect)
                                    if mergedAspect not in gStrctNode.aspects:
                                        if isinstance(aspect, list):
                                            aspect = set(aspect)
                                        gStrctNode.aspects[mergedAspect] = aspect
                                if childDefnMdlNode.tagSelector is not None and not gStrctNode.tagSelector:
                                    gStrctNode.tagSelector = childDefnMdlNode.tagSelector
                            # print(childStrctNode.aspectsCovered())
                        '''
                        childMergeAspects = mergeAspects.copy() if mergeAspects else {}
                        for mergedAspect in childDefnMdlNode.aspectsCovered():
                            aspect = childDefnMdlNode.aspectValue(view.rendrCntx, mergedAspect)
                            if isinstance(aspect, list):
                                aspect = set(aspect)
                            childMergeAspects[mergedAspect] = aspect
                        if childDefnMdlNode.tagSelector is not None:
                            childMergeAspects.setdefault("tagSelectors",set()).add(childDefnMdlNode.tagSelector)
                        childStrctNode = resolveDefinition(view, strctMdlNode, childDefnMdlNode, depth, facts, iBrkdn, axisBrkdnRels, mergeAspects=childMergeAspects)
                    elif defnMdlNode.isMerged and isinstance(childDefnMdlNode, (DefnMdlAspectNode,DefnMdlRelationshipNode)):
                        # parent is a merged node and child is an aspect node
                        childStrctNode = resolveDefinition(view, strctMdlNode, childDefnMdlNode, depth, facts, iBrkdn, axisBrkdnRels, mergeAspects=mergeAspects)
                    elif not isinstance(childDefnMdlNode, DefnMdlTable):
                        childStrctNode = resolveDefinition(view, strctMdlNode, childDefnMdlNode, depth+ordDepth, facts, iBrkdn, axisBrkdnRels)
                        for aspect in (mergeAspects or ()):
                            if aspect == "tagSelectors":
                                childStrctNode.tagSelectors |= mergeAspects["tagSelectors"]
                            elif aspect not in childStrctNode.aspectsCovered(inherit=False):
                                childStrctNode.aspects[aspect] = mergeAspects[aspect]
                        descendantDefMdlNodes = view.defnSubtreeRelSet.fromModelObject(childDefnMdlNode)
                        if not isinstance(childDefnMdlNode, DefnMdlAspectNode) and not childDefnMdlNode.isAbstract and descendantDefMdlNodes:
                            # contributes at least one child node
                            rollupAspectDefinitionNode = childDefnMdlNode
                            _rollup = ROLLUP_IMPLIES_DEFAULT_MEMBER
                            # definition node is that of the other child elements, to contribute a defaulted dimension
                            if childStrctNode.strctMdlChildNodes:
                                childStrctNodeAspectsCovered = childStrctNode.aspectsCovered()
                                if childStrctNodeAspectsCovered  & {Aspect.DIMENSIONS, Aspect.OMIT_DIMENSIONS}:
                                    if childStrctNodeAspectsCovered == childStrctNode.strctMdlChildNodes[0].aspectsCovered():
                                        _rollup = ROLLUP_SPECIFIES_MEMBER
                                    else:
                                        rollupAspectDefinitionNode = childStrctNode.strctMdlChildNodes[0].defnMdlNode
                            rollUpStrctNode = StrctMdlStructuralNode(childStrctNode, rollupAspectDefinitionNode)
                            rollUpStrctNode.rollup = _rollup
                            childStrctNode.hasChildRollup = True
                            childStrctNode.rollUpChildStrctMdlNode = rollUpStrctNode
                            if isinstance(childDefnMdlNode, DefnMdlClosedDefinitionNode) and childDefnMdlNode.parentChildOrder == "parent-first":
                                childStrctNode.strctMdlChildNodes = childStrctNode.strctMdlChildNodes[-1:] + childStrctNode.strctMdlChildNodes[0:-1]
                            cartesianProductExpander(rollUpStrctNode, *cartesianProductNestedArgs)
                        if not childContainsOpenNodes(childStrctNode) and not childDefnMdlNode.childrenCoverSameAspects:
                            # To be computed only if the structural node does not contain an open node
                            cartesianProductExpander(childStrctNode, *cartesianProductNestedArgs)
                        if childStrctNode is not None and childStrctNode.isUnreported and isinstance(childDefnMdlNode, DefnMdlOpenDefinitionNode):
                            hasUnreportedChildStrctNode = True
                            strctMdlNode.strctMdlChildNodes.remove(childStrctNode)
                if hasUnreportedChildStrctNode and not strctMdlNode.strctMdlChildNodes and childStrctNode is not None:
                    # has all child nodes unreported, move group for header up to this level
                    strctMdlNode.strctMdlChildNodes.extend(childStrctNode.strctMdlChildNodes)
                # check if a children specify explicit dimensions and one is missing default
                descendantDefMdlNodes = view.defnSubtreeRelSet.fromModelObject(defnMdlNode)
                defnMdlChildDefDimsCovered = set( # defaulted dims in children
                    aspect
                    for rel in descendantDefMdlNodes
                    if not isinstance(rel.toModelObject, (NoneType,DefnMdlTable))
                    for aspect in rel.toModelObject.aspectsCovered()
                    if isinstance(aspect, QName) and aspect in view.modelXbrl.qnameDimensionDefaults
                )
                # note child defnMdlNodes needing default dimension
                for rel in descendantDefMdlNodes:
                    descDefnMdlNode = rel.toModelObject
                    if not isinstance(descDefnMdlNode, (NoneType,DefnMdlTable)):
                        defaultedDims = defnMdlChildDefDimsCovered - descDefnMdlNode.aspectsCovered() - defnMdlNode.aspectsCovered()
                        if defaultedDims:
                            descDefnMdlNode.deemedDefaultedDims = defaultedDims
                # check if child strct nodes specify explicit dimensions and one is missing default
                strctMdlChildDefDimsCovered = set( # defaulted dims in children
                    aspect
                    for gStrctNode in strctMdlParent.strctMdlChildNodes
                    for aspect in gStrctNode.aspectsCovered(inherit=True)
                    if isinstance(aspect, QName) and aspect in view.modelXbrl.qnameDimensionDefaults and gStrctNode.axis == axis
                )
                # note child defnMdlNodes needing default dimension
                anyUncoveredChilds = (defnMdlChildDefDimsCovered and
                                      any(defnMdlChildDefDimsCovered - gStrctNode.aspectsCovered(inherit=True)
                                          for gStrctNode in strctMdlParent.strctMdlChildNodes))
                for gStrctNode in strctMdlParent.strctMdlChildNodes:
                    if gStrctNode.axis == axis: # don't cross over breakdowns of other axis
                        defaultedDims = strctMdlChildDefDimsCovered - gStrctNode.aspectsCovered(inherit=True)
                        # remove any OMIT dimensions
                        if gStrctNode.hasAspect(Aspect.OMIT_DIMENSIONS):
                            defaultedDims |= set(gStrctNode.aspectValue(Aspect.OMIT_DIMENSIONS))
                        if defaultedDims:
                            gStrctNode.defnMdlNode.deemedDefaultedDims = defaultedDims
                        elif (anyUncoveredChilds and gStrctNode.defnMdlNode is not None
                              and (defnMdlChildDefDimsCovered - gStrctNode.aspectsCovered(inherit=True))):
                            gStrctNode.defnMdlNode.deemedDefaultedDims = defnMdlChildDefDimsCovered
            if isinstance(defnMdlNode, DefnMdlRelationshipNode):
                rels = defnMdlNode.relationships(strctMdlParent)
                if defnMdlNode.isOrSelfAxis:
                    rootOrSelfStructuralNodes = {}
                    addRelationships(defnMdlNode, defnMdlNode._sourceQnames, strctMdlParent, rootOrSelfStructuralNodes)
                else:
                    rootOrSelfStructuralNodes = None
                addRelationships(defnMdlNode, rels, strctMdlParent, rootOrSelfStructuralNodes)
                trimAbstractNodes(strctMdlParent)
                addDescendantRollups(strctMdlParent)
                # set up by defnMdlNode.relationships
                if isinstance(defnMdlNode, DefnMdlConceptRelationshipNode):
                    if (defnMdlNode._sourceQnames != [XbrlConst.qnXfiRoot] and
                        any(c is None or c.isHypercubeItem or c.isDimensionItem
                            for qn in defnMdlNode._sourceQnames
                            for c in (view.modelXbrl.qnameConcepts.get(qn),))):
                        view.modelXbrl.error("xbrlte:invalidConceptRelationshipSource",
                            _("Concept relationship rule node %(xlinkLabel)s source %(source)s does not refer to an existing concept."),
                            modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel, source=", ".join(str(q) for q in defnMdlNode._sourceQnames))
                    else:
                        # check if single network
                        networks = set() # entry is linkrole, arcrole, linkQName, arcQName
                        for rel in flattenSequence(defnMdlNode.relationships(strctMdlNode)):
                            if isinstance(rel, ModelRelationship):
                                networks.add((rel.arcrole, rel.linkrole, rel.qname, rel.arcElement.qname))
                            else:
                                networks.add(None) # root element
                        if len(networks) > 1:
                            view.modelXbrl.error("xbrlte:ambiguousConceptNetwork",
                                _("Concept relationship rule node %(xlinkLabel)s has %(count)s networks: %(networks)s"),
                                modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel, count=len(networks), networks=str(networks))
                elif isinstance(defnMdlNode, DefnMdlDimensionRelationshipNode):
                    dim = view.modelXbrl.qnameConcepts.get(defnMdlNode._dimensionQname)
                    if (dim is None or not dim.isExplicitDimension): #  and len(view.modelXbrl.factsInInstance) > 0:
                        view.modelXbrl.error("xbrlte:invalidExplicitDimensionQName",
                            _("Dimension relationship rule node %(xlinkLabel)s dimension %(dimension)s does not refer to an existing explicit dimension."),
                            modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel, dimension=defnMdlNode._dimensionQname)
                    for _sourceQname in defnMdlNode._sourceQnames:
                        domMbr = view.modelXbrl.qnameConcepts.get(_sourceQname)
                        if domMbr is None or not domMbr.isDomainMember:
                            view.modelXbrl.error("xbrlte:invalidDimensionRelationshipSource",
                                _("Dimension relationship rule node %(xlinkLabel)s source %(source)s does not refer to an existing domain member."),
                                modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel, source=_sourceQname)
                if (defnMdlNode._formulaAxis in ("child", "child-or-self", "parent", "parent-or-self", "sibling", "sibling-or-self") and
                    (not isinstance(defnMdlNode._generations, Number) or defnMdlNode._generations > 1)):
                    view.modelXbrl.error("xbrlte:relationshipNodeTooManyGenerations ",
                        _("Relationship rule node %(xlinkLabel)s formulaAxis %(axis)s implies a single generation tree walk but generations %(generations)s is greater than one."),
                        modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel, axis=defnMdlNode._formulaAxis, generations=defnMdlNode._generations)
                for childStrctNode in strctMdlParent.strctMdlChildNodes:
                    for aspect in (mergeAspects or ()):
                        if aspect == "tagSelectors":
                            childStrctNode.tagSelectors |= mergeAspects["tagSelectors"]
                        elif not defnMdlNode.hasAspect(childStrctNode, aspect):
                            childStrctNode.aspects[aspect] = mergeAspects[aspect]

            elif isinstance(defnMdlNode, DefnMdlAspectNode):
                strctMdlNode.setHasOpenNode()
                strctMdlNode.isLabeled = False
                isCartesianProductExpanded = True

                # strctMdlNode.abstract = True # spanning ordinate acts as a subtitle
                aspectFactsPartitions = strctMdlNode.evaluate(defnMdlNode,
                                                                  defnMdlNode.filteredFactsPartitions,
                                                                  evalArgs=(view.modelXbrl.factsInInstance,)) # for all reported facts ignoring parent def node filters
                if depth < 2:
                    filteredFactsPartitions = aspectFactsPartitions
                else:
                    filteredFactsPartitions = strctMdlNode.evaluate(defnMdlNode,
                                                                  defnMdlNode.filteredFactsPartitions,
                                                                  evalArgs=(facts,))

                # apply mergedAspects

                if strctMdlNode._rendrCntx.formulaOptions.traceVariableFilterWinnowing:
                    view.modelXbrl.info("table:trace",
                        _("Filter node %(xlinkLabel)s facts partitions: %(factsPartitions)s"),
                        modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel,
                        factsPartitions=str(filteredFactsPartitions))
                # only for fact entry (true if no parent open nodes or all are on entry prototype row)
                childList = strctMdlNode.strctMdlChildNodes
                if strctMdlNode.isEntryPrototype(default=True):
                    for i in range(getattr(view, "openBreakdownLines",
                                           # for file output, 1 entry row if no facts
                                           0 if filteredFactsPartitions else 1)):
                        view.aspectEntryObjectId += 1
                        filteredFactsPartitions.append([FactPrototype(view, {"aspectEntryObjectId": OPEN_ASPECT_ENTRY_SURROGATE + str(view.aspectEntryObjectId)})])
                        if strctMdlNode.isEntryPrototype(default=False):
                            break # only one node per cartesian product under outermost nested open entry row
                if depth >= 2:
                    for i in range(len(filteredFactsPartitions)):
                        filteredFactsPartitions[i] = set(filteredFactsPartitions[i])
                for aspectPartition in aspectFactsPartitions:
                    childStrctMdlNode = StrctMdlStructuralNode(strctMdlNode, defnMdlNode, contextItemFact=aspectPartition[0])
                    aspectPartition = set(aspectPartition)

                    # find matching filterFactsPartition
                    if depth < 2:
                        factsPartition = aspectPartition
                    else:
                        factsPartition = set()
                        for fp in filteredFactsPartitions:
                            if fp <= aspectPartition:
                                factsPartition = fp
                                break

                    # store the partition for later reuse when spreading facts in body cells if no aspect filters involved
                    if not defnMdlNode.filterRelationships:
                        childStrctMdlNode.factsPartition = factsPartition

                    childStrctMdlNode.indent = 0
                    #TBD this is now computed, not an attribute
                    #childStrctMdlNode.depth -= 1  # for label width; parent is merged/invisible
                    checkLabelWidth(view, childStrctMdlNode, subtreeRels, checkBoundFact=True)
                    #resolveDefinition(view, childStrctMdlNode, breakdownNode, defnMdlNode, depth, axis, factsPartition, processOpenDefinitionNode=False) #recurse

                    for aspect in (mergeAspects or ()):
                        if aspect == "tagSelectors":
                            childStrctMdlNode.tagSelectors |= mergeAspects["tagSelectors"]
                        elif not defnMdlNode.hasAspect(childStrctMdlNode, aspect):
                            childStrctMdlNode.aspects[aspect] = mergeAspects[aspect]

                    if subtreeRels:
                        for subtreeRel in subtreeRels:
                            child2DefinitionNode = subtreeRel.toModelObject
                            #child2StructuralNode = StrctMdlStructuralNode(childStrctMdlNode, child2DefinitionNode) # others are nested structuralNode
                            #childStrctMdlNode.strctMdlChildNodes.append(child2StructuralNode)
                            #resolveDefinition(view, child2StructuralNode, child2DefinitionNode, depth+ordDepth, factsPartition, iBrkdn, axisBrkdnRels) #recurse
                            resolveDefinition(view, childStrctMdlNode, child2DefinitionNode, depth+ordDepth, factsPartition, iBrkdn, axisBrkdnRels) #recurse
                # sort by header (which is likely to be typed dim value, for example)
                childList.sort(key=lambda childStrctMdlNode:
                               childStrctMdlNode.header(lang=view.lang,
                                                          returnGenLabel=False,
                                                          returnMsgFormatString=False,
                                                          layoutMdlSortOrder=True)
                               or '') # exception on trying to sort if header returns None
                #test case 1200 v10i may require fact in instance order
                if defnMdlNode.aspectsCovered() == {Aspect.CONCEPT}:
                    childList.sort(key=lambda childStrctMdlNode:
                                   min(f.objectIndex for f in childStrctMdlNode.factsPartition) if getattr(childStrctMdlNode,"factsPartition",None) else 999999999)
                # TBD if there is no abstract 'sub header' for these subOrdCntxs, move them in place of parent structuralNode
            elif isinstance(defnMdlNode, DefnMdlRuleDefinitionNode):
                for constraintSet in defnMdlNode.constraintSets.values():
                    _aspectsCovered = constraintSet.aspectsCovered()
                    for aspect in _aspectsCovered:
                        if not constraintSet.aspectValueDependsOnVars(aspect):
                            if aspect == Aspect.CONCEPT:
                                conceptQname = constraintSet.aspectValue(view.rendrCntx, Aspect.CONCEPT)
                                concept = view.modelXbrl.qnameConcepts.get(conceptQname)
                                if concept is None or not concept.isItem or concept.isDimensionItem or concept.isHypercubeItem:
                                    view.modelXbrl.error("xbrlte:invalidQNameAspectValue",
                                        _("Rule node %(xlinkLabel)s specifies concept %(concept)s does not refer to an existing primary item concept."),
                                        modelObject=constraintSet, xlinkLabel=defnMdlNode.xlinkLabel, concept=conceptQname)
                            elif isinstance(aspect, QName):
                                dimConcept = view.modelXbrl.qnameConcepts.get(aspect)
                                dimVal = defnMdlNode.aspectValue(view.rendrCntx, aspect)
                                if dimConcept is not None:
                                    if dimConcept.isTypedDimension: # dimValue is the member node
                                        if isinstance(dimVal, ModelObject) and not view.modelXbrl.factsByDimMemQname(aspect, dimVal.textValue):
                                            strctMdlNode.isUnreported = True # typed Dim is unreported
                                    else: # dimValue is a QName
                                        mem = view.modelXbrl.qnameConcepts.get(dimVal)
                                        if isinstance(dimVal, QName) and (mem is None or not mem.isDomainMember) and dimVal != XbrlConst.qnFormulaDimensionSAV: # SAV is absent dimension member, reported in validateFormula:
                                            view.modelXbrl.error("xbrlte:invalidQNameAspectValue",
                                                _("Rule node %(xlinkLabel)s specifies domain member %(concept)s does not refer to an existing domain member concept."),
                                                modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel, concept=dimVal)
                #if not defnMdlNode.constraintSets:
                #    view.modelXbrl.error("xbrlte:incompleteAspectRule",
                #        _("Rule node %(xlinkLabel)s does not specify an aspect value."),
                #        modelObject=defnMdlNode, xlinkLabel=defnMdlNode.xlinkLabel)
            if not isCartesianProductExpanded:
                cartesianProductExpander(strctMdlNode, *cartesianProductNestedArgs)

            if isinstance(defnMdlNode, (NoneType, DefnMdlBreakdown)) and not strctMdlNode.strctMdlChildNodes: # childless root ordinate, make a child to iterate in producing table
                subOrdContext = StrctMdlBreakdown(strctMdlNode, defnMdlNode, axis)
        except ResolutionException as ex:
            if sys.version[0] >= '3':
                #import traceback
                #traceback.print_tb(ex.__traceback__)
                raise ex.with_traceback(ex.__traceback__)  # provide original traceback information
            else:
                raise ex
        except Exception as ex:
            e = ResolutionException("arelle:resolutionException",
                                    _("Exception in resolution of definition node %(node)s: %(error)s"),
                                    modelObject=defnMdlNode, node=defnMdlNode.qname, error=str(ex)
                                    )
            if sys.version[0] >= '3':
                raise e.with_traceback(ex.__traceback__)  # provide original traceback information
            else:
                raise e

    elif strctMdlNode and defnMdlNode is None: # no breakdown nodes for axis
        cartesianProductNestedArgs = [view, depth+1, axis, facts, (), iBrkdn]
        cartesianProductExpander(strctMdlNode, *cartesianProductNestedArgs)

    return strctMdlNode

def cartesianProductExpander(childStrctMdlNode, view, depth, axis, facts, axisBrkdnRels, iBrkdn):
    if iBrkdn is not None: # recurse table relationships for cartesian product
        for j, axisBrkdnRel in enumerate(axisBrkdnRels[iBrkdn+1:]):
            brkdnDefnMdlNode = axisBrkdnRel.toModelObject
            if isinstance(brkdnDefnMdlNode, DefnMdlBreakdown):
                subOrdTblCntx = childStrctMdlNode
                # predefined axes need facts sub-filtered
                if isinstance(childStrctMdlNode.defnMdlNode, DefnMdlClosedDefinitionNode):
                    matchingFacts = childStrctMdlNode.evaluate(childStrctMdlNode.defnMdlNode,
                                                        childStrctMdlNode.defnMdlNode.filteredFacts,
                                                        evalArgs=(facts,))
                else:
                    matchingFacts = facts
                # returns whether there were no structural node results
                subOrdTblCntx.abstract = True # can't be abstract across breakdown
                resolveDefinition(view, subOrdTblCntx, brkdnDefnMdlNode,
                            depth, # depth + (0 if axis == 'z' else 1),
                            matchingFacts, j + iBrkdn + 1, axisBrkdnRels,
                            axis=subOrdTblCntx.axis) #cartesian product
                break

def addRelationship(relDefinitionNode, rel, strctMdlNode, rootOrSelfStructuralNodes=None):
    variableQname = relDefinitionNode.variableQname
    conceptQname = relDefinitionNode.conceptQname
    coveredAspect = relDefinitionNode.coveredAspect(strctMdlNode)
    if not coveredAspect:
        return None
    if rootOrSelfStructuralNodes is not None:
        if isinstance(rel, QName):
            fromConceptQname = rel
            if fromConceptQname == XbrlConst.qnXfiRoot:
                return None
        else:
            fromConceptQname = rel.fromModelObject.qname
        # is there an ordinate for this root object?
        if fromConceptQname in rootOrSelfStructuralNodes:
            childStrctMdlNode = rootOrSelfStructuralNodes[fromConceptQname]
        else:
            childStrctMdlNode = StrctMdlStructuralNode(strctMdlNode, relDefinitionNode)
            rootOrSelfStructuralNodes[fromConceptQname] = childStrctMdlNode
            if variableQname:
                childStrctMdlNode.variables[variableQname] = []
            if conceptQname:
                childStrctMdlNode.variables[conceptQname] = fromConceptQname
            childStrctMdlNode.aspects[coveredAspect] = fromConceptQname
            concept = relDefinitionNode.modelXbrl.qnameConcepts.get(fromConceptQname)
            if concept is not None and concept.isAbstract:
                childStrctMdlNode.abstract = True
        if isinstance(rel, QName):
            return childStrctMdlNode
        relChildStrctMdlNode = StrctMdlStructuralNode(childStrctMdlNode, relDefinitionNode)
    else:
        relChildStrctMdlNode = StrctMdlStructuralNode(strctMdlNode, relDefinitionNode)
    if isinstance(rel, ModelRelationship):
        if isinstance(relDefinitionNode, DefnMdlConceptRelationshipNode):
            preferredLabel = rel.preferredLabel
            genPreferredLabel = rel.get("{http://xbrl.org/2013/preferred-label}preferredLabel")
            if preferredLabel is None:
                if genPreferredLabel is not None:
                    preferredLabel = genPreferredLabel
            elif genPreferredLabel is not None:
                rel.modelXbrl.error("xbrlte:ambiguousPreferredLabel",
                    _("PresentationArc has both the preferredLabel (%(preferredLabel)s) and gpl:preferredLabel (%(genPreferredLabel)s) attributes"),
                    modelObject=rel, preferredLabel=preferredLabel, genPreferredLabel=genPreferredLabel)
            if preferredLabel == XbrlConst.periodStartLabel:
                relChildStrctMdlNode.tagSelector = "table.periodStart"
            elif preferredLabel == XbrlConst.periodEndLabel:
                relChildStrctMdlNode.tagSelector = "table.periodEnd"
        elif isinstance(relDefinitionNode, DefnMdlDimensionRelationshipNode):
            relChildStrctMdlNode.abstract = not rel.isUsable
        toConceptQname = rel.toModelObject.qname
    else:
        toConceptQname = rel # QName
    if variableQname:
        relChildStrctMdlNode.variables[variableQname] = rel
    if conceptQname:
        relChildStrctMdlNode.variables[conceptQname] = toConceptQname
    relChildStrctMdlNode.aspects[coveredAspect] = toConceptQname
    concept = relDefinitionNode.modelXbrl.qnameConcepts.get(toConceptQname)
    if isinstance(relDefinitionNode, DefnMdlConceptRelationshipNode) and concept is not None and concept.isAbstract:
        relChildStrctMdlNode.abstract = True
    return relChildStrctMdlNode

def addRelationships(relDefinitionNode, rels, strctMdlNode, rootOrSelfStructuralNodes=None):
    childStrctMdlNode = None # holder for nested relationships
    for rel in rels:
        if not isinstance(rel, list):
            # first entry can be parent of nested list relationships
            childStrctMdlNode = addRelationship(relDefinitionNode, rel, strctMdlNode, rootOrSelfStructuralNodes)
        elif childStrctMdlNode is None:
            childStrctMdlNode = StrctMdlStructuralNode(strctMdlNode, relDefinitionNode)
            addRelationships(relDefinitionNode, rel, childStrctMdlNode)
        else:
            addRelationships(relDefinitionNode, rel, childStrctMdlNode)

def trimAbstractNodes(strctMdlParent):
    for childStrctMdlNode in strctMdlParent.strctMdlChildNodes:
        trimAbstractNodes(childStrctMdlNode)
    # remove childless abstract structural nodes by tail recursion
    abstractChildlessNodes = []
    for i, childStrctMdlNode in enumerate(strctMdlParent.strctMdlChildNodes):
        if childStrctMdlNode.isAbstract and not childStrctMdlNode.strctMdlChildNodes:
            abstractChildlessNodes.insert(0, i) # reversed list
    for i in abstractChildlessNodes:
        del strctMdlParent.strctMdlChildNodes[i]

def getDepth(strctMdlParent, depth=0):
    maxDepth = depth
    for childStrctMdlNode in strctMdlParent.strctMdlChildNodes:
        d = getDepth(childStrctMdlNode, depth+1)
        if d > maxDepth:
            maxDepth = d
    return maxDepth


def addDescendantRollups(strctMdlParent, maxDepth=None, depth=0):
    if maxDepth is None:
        maxDepth = getDepth(strctMdlParent)
    # add roll-up nodes after the children were populated from top level
    needsChildRollup = 0 < depth < maxDepth and (not strctMdlParent.isAbstract or strctMdlParent.rollup)
    for childStrctMdlNode in strctMdlParent.strctMdlChildNodes:
        addDescendantRollups(childStrctMdlNode, maxDepth, depth=depth+1)
    if needsChildRollup:
        rollUpStrctNode = StrctMdlStructuralNode(strctMdlParent, strctMdlParent.defnMdlNode)
        rollUpStrctNode.rollup = strctMdlParent.defnMdlNode.strctMdlRollupType
        strctMdlParent.hasChildRollup = True
        strctMdlParent.rollUpChildStrctMdlNode = rollUpStrctNode
        if len(strctMdlParent.strctMdlChildNodes) > 1 and strctMdlParent.parentChildOrder == "parent-first":
            strctMdlParent.strctMdlChildNodes = strctMdlParent.strctMdlChildNodes[-1:] + strctMdlParent.strctMdlChildNodes[0:-1]
        addDescendantRollups(rollUpStrctNode, maxDepth, depth=depth+1)

def addDefaultedDimensionsToLeafNodes(strctMdlNode, coveredDims=None, defaultedDims=None):
    if coveredDims is None:
        coveredDims = set()
        defaultedDims = set()
    coveredDims = coveredDims | strctMdlNode.aspectsCovered()
    if strctMdlNode.hasAspect(Aspect.OMIT_DIMENSIONS):
        coveredDims = coveredDims - set(strctMdlNode.aspectValue(Aspect.OMIT_DIMENSIONS))
    if hasattr(strctMdlNode.defnMdlNode, "deemedDefaultedDims"):
        defaultedDims = defaultedDims | getattr(strctMdlNode.defnMdlNode, "deemedDefaultedDims")
    if strctMdlNode.rollup == ROLLUP_IMPLIES_DEFAULT_MEMBER:
        deemedDefaultedDims = defaultedDims - coveredDims
        if deemedDefaultedDims:
            strctMdlNode.deemedDefaultedDims = deemedDefaultedDims
            defaultedDims -= deemedDefaultedDims
    if strctMdlNode.strctMdlChildNodes:
        for childStrctMdlNode in strctMdlNode.strctMdlChildNodes:
            addDefaultedDimensionsToLeafNodes(childStrctMdlNode, coveredDims, defaultedDims)
    else: # leaf node
        deemedDefaultedDims = defaultedDims - coveredDims
        if deemedDefaultedDims:
            strctMdlNode.deemedDefaultedDims = deemedDefaultedDims



"""
See COPYRIGHT.md for copyright information.
"""
from __future__ import annotations

from typing import Any

from arelle.Cntlr import Cntlr
from arelle.FileSource import FileSource
from arelle.ValidateXbrl import ValidateXbrl
from arelle.typing import TypeGetText
from arelle.utils.PluginData import PluginData
from arelle.utils.validate.ValidationPlugin import ValidationPlugin
from .ControllerPluginData import ControllerPluginData
from .DisclosureSystems import DISCLOSURE_SYSTEM_EDINET
from .ManifestInstance import parseManifests
from .PluginValidationDataExtension import PluginValidationDataExtension

_: TypeGetText


class ValidationPluginExtension(ValidationPlugin):

    def fileSourceEntrypointFiles(self, filesource: FileSource, *args: Any, **kwargs: Any) -> list[dict[str, Any]] | None:
        instances = parseManifests(filesource)
        if len(instances) == 0:
            return None
        assert filesource.cntlr is not None
        pluginData = ControllerPluginData.get(filesource.cntlr, self.name)
        entrypointFiles = []
        for instance in instances:
            pluginData.addManifestInstance(instance)
            entrypoints = []
            for ixbrlFile in instance.ixbrlFiles:
                filesource.select(str(ixbrlFile))
                entrypoints.append({"file": filesource.url})
            entrypointFiles.append({'ixds': entrypoints, 'id': instance.id})
        return entrypointFiles

    def newPluginData(self, cntlr: Cntlr, validateXbrl: ValidateXbrl | None) -> PluginData:
        if validateXbrl is None:
            return ControllerPluginData.get(cntlr, self.name)
        disclosureSystem = DISCLOSURE_SYSTEM_EDINET
        if validateXbrl is not None:
            disclosureSystem = str(validateXbrl.disclosureSystem.name)
        if disclosureSystem == DISCLOSURE_SYSTEM_EDINET:
            pass
        else:
            raise ValueError(f'Invalid EDINET disclosure system: {disclosureSystem}')
        return PluginValidationDataExtension(
            self.name,
        )



'''
XbrlSemanticSqlDB.py implements an SQL database interface for Arelle, based
on a concrete realization of the Abstract Model PWD 2.0 layer.  This is a semantic
representation of XBRL information.

This module may save directly to a Postgres, MySQL, SQLite, MSSQL, or Oracle server.

This module provides the execution context for saving a dts and instances in
XBRL SQL database.  It may be loaded by Arelle's RSS feed, or by individual
DTS and instances opened by interactive or command line/web service mode.

Example dialog or command line parameters for operation:

    host:  the supporting host for SQL Server
    port:  the host port of server
    user, password:  if needed for server
    database:  the top level path segment for the SQL Server
    timeout:


See COPYRIGHT.md for copyright information.


to use from command line:

linux
   # be sure plugin is installed
   arelleCmdLine --plugin '+xbrlDB|show'
   arelleCmdLine -f http://sec.org/somewhere/some.rss -v --store-to-XBRL-DB 'myserver.com,portnumber,pguser,pgpasswd,database,timeoutseconds'

windows
   rem be sure plugin is installed
   arelleCmdLine --plugin "xbrlDB"
   arelleCmdLine -f http://sec.org/somewhere/some.rss -v --store-to-XBRL-DB "myserver.com,portnumber,pguser,pgpasswd,database,timeoutseconds"

'''

import os, time, datetime, logging
from arelle.ModelDocument import Type
from arelle.ModelDtsObject import ModelConcept, ModelType, ModelResource, ModelRelationship
from arelle.ModelInstanceObject import ModelFact
from arelle.ModelXbrl import ModelXbrl
from arelle.ModelDocument import ModelDocument
from arelle.ModelObject import ModelObject
from arelle.ModelValue import qname
from arelle.ValidateXbrlCalcs import roundValue
from arelle.XmlValidate import UNVALIDATED, VALID
from arelle.XmlUtil import elementChildSequence
from arelle import XbrlConst
from arelle.UrlUtil import authority, ensureUrl
from .SqlDb import XPDBException, isSqlConnection, SqlDbConnection
from .tableFacts import tableFacts
from .entityInformation import loadEntityInformation
from .primaryDocumentFacts import loadPrimaryDocumentFacts
from collections import defaultdict


def insertIntoDB(modelXbrl,
                 user=None, password=None, host=None, port=None, database=None, timeout=None,
                 product=None, entrypoint=None, rssItem=None, **kwargs):
    xbrlDbConn = None
    try:
        xbrlDbConn = XbrlSqlDatabaseConnection(modelXbrl, user, password, host, port, database, timeout, product)
        if "rssObject" in kwargs: # initialize batch
            xbrlDbConn.initializeBatch(kwargs["rssObject"])
        else:
            xbrlDbConn.verifyTables()
            xbrlDbConn.insertXbrl(entrypoint, rssItem)
        xbrlDbConn.close()
    except Exception as ex:
        if xbrlDbConn is not None:
            try:
                xbrlDbConn.close(rollback=True)
            except Exception as ex2:
                pass
        raise # reraise original exception with original traceback

def isDBPort(host, port, timeout=10, product="postgres"):
    return isSqlConnection(host, port, timeout)

XBRLDBTABLES = {
                "filing", "report",
                "document", "referenced_documents",
                "aspect", "data_type", "role_type", "arcrole_type",
                "resource", "relationship_set", "root", "relationship",
                "data_point", "entity", "period", "unit", "unit_measure", "aspect_value_selection",
                "message", "message_reference",
                "industry", "industry_level", "industry_structure",
                }



class XbrlSqlDatabaseConnection(SqlDbConnection):
    def verifyTables(self):
        missingTables = XBRLDBTABLES - self.tablesInDB()
        # if no tables, initialize database
        if missingTables == XBRLDBTABLES:
            self.create(os.path.join("sql", "semantic", {"mssql": "xbrlSemanticMSSqlDB.sql",
                                                         "mysql": "xbrlSemanticMySqlDB.ddl",
                                                         "sqlite": "xbrlSemanticSQLiteDB.ddl",
                                                         "orcl": "xbrlSemanticOracleDB.sql",
                                                         "postgres": "xbrlSemanticPostgresDB.ddl"}[self.product]))
            missingTables = XBRLDBTABLES - self.tablesInDB()
        if missingTables and missingTables != {"sequences"}:
            raise XPDBException("sqlDB:MissingTables",
                                _("The following tables are missing: %(missingTableNames)s"),
                                missingTableNames=', '.join(t for t in sorted(missingTables)))

    def insertXbrl(self, entrypoint, rssItem):
        try:
            # must also have default dimensions loaded
            from arelle import ValidateXbrlDimensions
            ValidateXbrlDimensions.loadDimensionDefaults(self.modelXbrl)

            # get logging entries (needed to find which aspects to identify)
            self.loggingEntries = []
            for handler in logging.getLogger("arelle").handlers:
                if hasattr(handler, "dbHandlerLogEntries"):
                    self.loggingEntries = handler.dbHandlerLogEntries()
                    break

            # must have a valid XBRL instance or document
            if self.modelXbrl.modelDocument is None:
                raise XPDBException("xpgDB:MissingXbrlDocument",
                                    _("No XBRL instance or schema loaded for this filing."))

            # obtain supplementaion entity information
            self.entityInformation = loadEntityInformation(self.modelXbrl, entrypoint, rssItem)
            # identify table facts (table datapoints) (prior to locked database transaction
            self.tableFacts = tableFacts(self.modelXbrl)  # for EFM & HMRC this is ( (roleType, table_code, fact) )
            loadPrimaryDocumentFacts(self.modelXbrl, rssItem, self.entityInformation) # load primary document facts for SEC filing
            self.identifyTaxonomyRelSetsOwner()

            # at this point we determine what's in the database and provide new tables
            # requires locking most of the table structure
            self.lockTables(('entity', 'filing', 'report', 'document', 'referenced_documents'),
                            isSessionTransaction=True) # lock for whole transaction

            # find pre-existing documents in server database
            self.identifyPreexistingDocuments()
            self.identifyAspectsUsed()

            self.dropTemporaryTable()
            startedAt = time.time()
            self.syncSequences = True # for data base types that don't explicity handle sequences
            self.insertFiling(rssItem)
            self.modelXbrl.profileStat(_("XbrlSqlDB: Filing insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertDocuments()
            self.modelXbrl.profileStat(_("XbrlSqlDB: Documents insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertAspects()
            self.modelXbrl.profileStat(_("XbrlSqlDB: Aspects insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertArcroleTypes()
            self.insertRoleTypes()
            self.modelXbrl.profileStat(_("XbrlSqlDB: Role Types insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertResources()
            self.modelXbrl.profileStat(_("XbrlSqlDB: Resources insertion"), time.time() - startedAt)
            startedAt = time.time()
            # self.modelXbrl.profileStat(_("XbrlSqlDB: DTS insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertDataPoints()
            self.modelXbrl.profileStat(_("XbrlSqlDB: instance insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertRelationships() # must follow data points for footnote relationships
            self.modelXbrl.profileStat(_("XbrlSqlDB: Relationships insertion"), time.time() - startedAt)
            startedAt = time.time()
            self.insertValidationResults()
            self.modelXbrl.profileStat(_("XbrlSqlDB: Validation results insertion"), time.time() - startedAt)

            startedAt = time.time()
            self.showStatus("Committing entries")
            self.commit()
            self.modelXbrl.profileStat(_("XbrlSqlDB: insertion committed"), time.time() - startedAt)
            self.showStatus("DB insertion completed", clearAfter=5000)
        except Exception as ex:
            self.showStatus("DB insertion failed due to exception", clearAfter=5000)
            raise

    def identifyTaxonomyRelSetsOwner(self):
        # walk down referenced document set from instance to find 'lowest' taxonomy relationship set ownership
        instanceReferencedDocuments = set()
        instanceDocuments = set()
        inlineXbrlDocSet = None
        for mdlDoc in self.modelXbrl.urlDocs.values():
            if mdlDoc.type in (Type.INSTANCE, Type.INLINEXBRL):
                instanceDocuments.add(mdlDoc)
                for refDoc, ref in mdlDoc.referencesDocument.items():
                    if refDoc.inDTS and ref.referenceTypes & {"href", "import", "include"}:
                        instanceReferencedDocuments.add(refDoc)
            elif mdlDoc.type == Type.INLINEXBRLDOCUMENTSET:
                inlineXbrlDocSet = mdlDoc
        if len(instanceReferencedDocuments) > 1:
            # filing must own the taxonomy set
            if len(instanceDocuments) == 1:
                self.taxonomyRelSetsOwner = instanceDocuments.pop()
            elif inlineXbrlDocSet is not None:  # manifest for inline docs can own the rel sets
                self.taxonomyRelSetsOwner = inlineXbrlDocSet
            else: # no single instance, pick the entry poin doct
                self.taxonomyRelSetsOwner = self.modelXbrl.modelDocument # entry document (instance or inline doc set)
        elif len(instanceReferencedDocuments) == 1:
            self.taxonomyRelSetsOwner = instanceReferencedDocuments.pop()
        elif self.modelXbrl.modelDocument.type == Type.SCHEMA:
            self.taxonomyRelSetsOwner = self.modelXbrl.modelDocument
        else:
            self.taxonomyRelSetsOwner = self.modelXbrl.modelDocument
        instanceReferencedDocuments.clear() # dereference
        instanceDocuments.clear()

        # check whether relationship_set is completely in instance or part/all in taxonomy
        self.arcroleInInstance = {}
        self.arcroleHasResource = {}
        for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys():
            if ELR is None and linkqname is None and arcqname is None and not arcrole.startswith("XBRL-"):
                inInstance = False
                hasResource = False
                for rel in self.modelXbrl.relationshipSet(arcrole).modelRelationships:
                    if (not inInstance and
                        rel.modelDocument.type in (Type.INSTANCE, Type.INLINEXBRL) and
                        any(isinstance(tgtObj, ModelObject) and tgtObj.modelDocument.type in (Type.INSTANCE, Type.INLINEXBRL)
                            for tgtObj in (rel.fromModelObject, rel.toModelObject))):
                        inInstance = True
                    if not hasResource and any(isinstance(resource, ModelResource)
                                               for resource in (rel.fromModelObject, rel.toModelObject)):
                        hasResource = True
                    if inInstance and hasResource:
                        break;
                self.arcroleInInstance[arcrole] = inInstance
                self.arcroleHasResource[arcrole] = hasResource

    def initializeBatch(self, rssObject):
        results = self.execute("SELECT filing_number, accepted_timestamp FROM filing")
        existingFilings = dict((filingNumber, timestamp)
                               for filingNumber, timestamp in results) # timestamp is a string
        for rssItem in rssObject.rssItems:
            if (rssItem.accessionNumber in existingFilings and
                rssItem.acceptanceDatetime == existingFilings[rssItem.accessionNumber]):
                rssItem.skipRssItem = True


    def insertFiling(self, rssItem):
        now = datetime.datetime.now()
        entityInfo = self.entityInformation
        def rssItemGet(propertyName):
            if rssItem is not None:
                return getattr(rssItem, propertyName, None)
            return None
        self.showStatus("insert entity")
        LEI = None
        entity_comparator = ('legal_entity_number', 'file_number') if LEI else ('file_number',)
        table = self.getTable('entity', 'entity_id',
                              ('legal_entity_number',
                               'file_number',
                               'reference_number', # CIK
                               'tax_number',
                               'standard_industry_code',
                               'name',
                               'legal_state',
                               'phone',
                               'phys_addr1', 'phys_addr2', 'phys_city', 'phys_state', 'phys_zip', 'phys_country',
                               'mail_addr1', 'mail_addr2', 'mail_city', 'mail_state', 'mail_zip', 'mail_country',
                               'fiscal_year_end',
                               'filer_category',
                               'public_float',
                               'trading_symbol'),
                              entity_comparator, # cannot compare None = None if LEI is absent, always False
                              ((LEI,
                                rssItemGet("fileNumber") or entityInfo.get("file-number")  or str(int(time.time())),
                                rssItemGet("cikNumber") or entityInfo.get("cik"),
                                entityInfo.get("irs-number"),
                                rssItemGet("assignedSic") or entityInfo.get("assigned-sic") or -1,
                                rssItemGet("companyName") or entityInfo.get("conformed-name"),
                                entityInfo.get("state-of-incorporation"),
                                entityInfo.get("business-address.phone"),
                                entityInfo.get("business-address.street1"),
                                entityInfo.get("business-address.street2"),
                                entityInfo.get("business-address.city"),
                                entityInfo.get("business-address.state"),
                                entityInfo.get("business-address.zip"),
                                countryOfState.get(entityInfo.get("business-address.state")),
                                entityInfo.get("mail-address.street1"),
                                entityInfo.get("mail-address.street2"),
                                entityInfo.get("mail-address.city"),
                                entityInfo.get("mail-address.state"),
                                entityInfo.get("mail-address.zip"),
                                countryOfState.get(entityInfo.get("mail-address.state")),
                                rssItemGet("fiscalYearEnd") or entityInfo.get("fiscal-year-end"),
                                entityInfo.get("filer-category"),
                                entityInfo.get("public-float"),
                                entityInfo.get("trading-symbol")
                                ),),
                              checkIfExisting=True,
                              returnExistenceStatus=True)
        if LEI:
            for id, _LEI, filing_number, existenceStatus in table:
                self.entityId = id
                self.entityPreviouslyInDB = existenceStatus
                break
        else:
            for id, filing_number, existenceStatus in table:
                self.entityId = id
                self.entityPreviouslyInDB = existenceStatus
                break
        if any ('former-conformed-name' in key for key in entityInfo.keys()):
            self.getTable('former_entity', None,
                          ('entity_id', 'former_name', 'date_changed'),
                          ('entity_id', 'former_name', 'date_changed'),
                          ((self.entityId,
                            entityInfo.get(keyPrefix + '.former-conformed-name'),
                            entityInfo.get(keyPrefix + '.date-changed'))
                           for key in entityInfo.keys() if 'former-conformed-name' in key
                           for keyPrefix in (key.partition('.')[0],)),
                          checkIfExisting=True)
        self.showStatus("insert filing")
        table = self.getTable('filing', 'filing_id',
                              ('filing_number', 'form_type', 'entity_id', 'reference_number',
                               'accepted_timestamp', 'is_most_current', 'filing_date',
                               'creation_software',
                               'authority_html_url', 'entry_url', ),
                              ('filing_number',),
                              ((rssItemGet("accessionNumber") or entityInfo.get("accession-number") or str(int(time.time())),  # NOT NULL
                                rssItemGet("formType") or entityInfo.get("form-type"),
                                self.entityId,
                                rssItemGet("cikNumber") or entityInfo.get("cik"),
                                rssItemGet("acceptanceDatetime") or entityInfo.get("acceptance-datetime") or now,
                                True,
                                rssItemGet("filingDate") or entityInfo.get("filing-date") or now,  # NOT NULL
                                self.modelXbrl.modelDocument.creationSoftware,
                                rssItemGet("htmlUrl") or entityInfo.get("primary-document-url"),
                                rssItemGet("url") or entityInfo.get("instance-url")
                                ),),
                              checkIfExisting=True,
                              returnExistenceStatus=True)
        for id, filing_number, existenceStatus in table:
            self.filingId = id
            self.filingPreviouslyInDB = existenceStatus
            break
        self.showStatus("insert report")
        table = self.getTable('report', 'report_id',
                              ('filing_id', ),
                              ('filing_id',),
                              ((self.filingId,
                                ),),
                              checkIfExisting=True,
                              returnExistenceStatus=True)
        for id, foundFilingId, existenceStatus in table:
            self.reportId = id
            self.filingPreviouslyInDB = existenceStatus
            break

    def isSemanticDocument(self, modelDocument):
        if modelDocument.type == Type.SCHEMA:
            # must include document items taxonomy even if not in DTS
            return modelDocument.inDTS or modelDocument.targetNamespace == "http://arelle.org/doc/2014-01-31"
        return modelDocument.type in (Type.INSTANCE, Type.INLINEXBRL, Type.LINKBASE)

    def identifyPreexistingDocuments(self):
        self.existingDocumentIds = {}
        self.urlDocs = {}
        docUris = set()
        for modelDocument in self.modelXbrl.urlDocs.values():
            url = ensureUrl(modelDocument.uri)
            self.urlDocs[url] = modelDocument
            if self.isSemanticDocument(modelDocument):
                docUris.add(self.dbStr(url))
        if docUris:
            results = self.execute("SELECT document_id, document_url FROM {} WHERE document_url IN ({})"
                                   .format(self.dbTableName("document"),
                                           ', '.join(docUris)))
            self.existingDocumentIds = dict((self.urlDocs[self.pyStrFromDbStr(docUrl)],docId)
                                            for docId, docUrl in results)

            # identify whether taxonomyRelsSetsOwner is existing
            self.isExistingTaxonomyRelSetsOwner = (
                self.taxonomyRelSetsOwner.type not in (Type.INSTANCE, Type.INLINEXBRL, Type.INLINEXBRLDOCUMENTSET) and
                self.taxonomyRelSetsOwner in self.existingDocumentIds)

    def identifyAspectsUsed(self):
        # relationshipSets are a dts property
        self.relationshipSets = [(arcrole, ELR, linkqname, arcqname)
                                 for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys()
                                 if ELR and (arcrole.startswith("XBRL-") or (linkqname and arcqname))]


        aspectsUsed = set(f.concept
                          for f in self.modelXbrl.factsInInstance)

        for cntx in self.modelXbrl.contexts.values():
            for dim in cntx.qnameDims.values():
                aspectsUsed.add(dim.dimension)
                if dim.isExplicit:
                    aspectsUsed.add(dim.member)
                else:
                    aspectsUsed.add(self.modelXbrl.qnameConcepts[dim.typedMember.qname])
        for defaultDimQn, defaultDimMemberQn in self.modelXbrl.qnameDimensionDefaults.items():
            aspectsUsed.add(self.modelXbrl.qnameConcepts[defaultDimQn])
            aspectsUsed.add(self.modelXbrl.qnameConcepts[defaultDimMemberQn])
        for relationshipSetKey in self.relationshipSets:
            relationshipSet = self.modelXbrl.relationshipSet(*relationshipSetKey)
            for rel in relationshipSet.modelRelationships:
                if isinstance(rel.fromModelObject, ModelConcept):
                    aspectsUsed.add(rel.fromModelObject)
                if isinstance(rel.toModelObject, ModelConcept):
                    aspectsUsed.add(rel.toModelObject)

        try:
            for qn in (XbrlConst.qnXbrliIdentifier, XbrlConst.qnXbrliPeriod, XbrlConst.qnXbrliUnit):
                aspectsUsed.add(self.modelXbrl.qnameConcepts[qn])
        except KeyError:
            pass # no DTS

        for roleTypes in (self.modelXbrl.roleTypes.values(), self.modelXbrl.arcroleTypes.values()):
            for roleUriTypes in roleTypes:
                for roleType in roleUriTypes:
                    for qn in roleType.usedOns:
                        if qn in self.modelXbrl.qnameConcepts: # qname may be undefined or invalid and still 2.1 legal
                            aspectsUsed.add(self.modelXbrl.qnameConcepts[qn])

        # add aspects referenced by logging entries
        for logEntry in self.loggingEntries:
            for ref in logEntry['refs']:
                modelObject = self.modelXbrl.modelObject(ref.get('objectId',''))
                if isinstance(modelObject, ModelConcept) and modelObject.modelDocument.inDTS:
                    aspectsUsed.add(modelObject)

        # add substitution groups
        aspectsUsed |= set(aspect.substitutionGroup
                           for aspect in aspectsUsed
                           if aspect is not None)

        aspectsUsed -= {None}  # remove None if in aspectsUsed
        self.aspectsUsed = aspectsUsed

        typesUsed = set()
        def typeUsed(modelType):
            if modelType is not None and modelType.modelDocument.inDTS: # exclude nonDTS types (schema, etc)
                typesUsed.add(modelType)
                typesDerivedFrom = modelType.typeDerivedFrom
                if isinstance(typesDerivedFrom, list): # union derivation
                    for typeDerivedFrom in typesDerivedFrom:
                        if typeDerivedFrom not in typesUsed:
                            typeUsed(typeDerivedFrom)
                else: # single derivation
                    if typesDerivedFrom is not None and typesDerivedFrom not in typesUsed:
                        typeUsed(typesDerivedFrom)

        for aspect in aspectsUsed:
            modelType = aspect.type
            if modelType is not None:
                if modelType not in typesUsed:
                    typeUsed(modelType)
        self.typesUsed = typesUsed

    def insertDocuments(self):
        self.showStatus("insert documents")
        table = self.getTable('document', 'document_id',
                              ('document_url', 'document_type', 'namespace'),
                              ('document_url',),
                              set((ensureUrl(docUrl),
                                   Type.typeName[mdlDoc.type],
                                   mdlDoc.targetNamespace)
                                  for docUrl, mdlDoc in self.modelXbrl.urlDocs.items()
                                  if mdlDoc not in self.existingDocumentIds and
                                     self.isSemanticDocument(mdlDoc)),
                              checkIfExisting=True)
        self.documentIds = dict((self.urlDocs[self.pyStrFromDbStr(url)], id)
                                for id, url in table)
        self.documentIds.update(self.existingDocumentIds)

        referencedDocuments = set()
        # instance documents are filing references
        # update report with document references
        for mdlDoc in self.modelXbrl.urlDocs.values():
            if mdlDoc.type in (Type.INSTANCE, Type.INLINEXBRL):
                referencedDocuments.add( (self.reportId, self.documentIds[mdlDoc] ))
            if mdlDoc in self.documentIds:
                for refDoc, ref in mdlDoc.referencesDocument.items():
                    if refDoc.inDTS and ref.referenceTypes & {"href", "import", "include"} \
                       and refDoc in self.documentIds:
                        referencedDocuments.add( (self.documentIds[mdlDoc], self.documentIds[refDoc] ))

        table = self.getTable('referenced_documents',
                              None, # no id column in this table
                              ('object_id','document_id'),
                              ('object_id','document_id'),
                              referencedDocuments,
                              checkIfExisting=True)

        instDocId = instSchemaDocId = agencySchemaDocId = stdSchemaDocId = None
        mdlDoc = self.modelXbrl.modelDocument
        if mdlDoc.type in (Type.INSTANCE, Type.INLINEXBRL):
            instDocId = self.documentIds[mdlDoc]
            # referenced doc may be extension schema
            for refDoc, ref in mdlDoc.referencesDocument.items():
                if refDoc.inDTS and "href" in ref.referenceTypes and refDoc in self.documentIds:
                    instSchemaDocId = self.documentIds[refDoc]
                    break
        elif mdlDoc.type == Type.SCHEMA:
            instDocSchemaDocId = self.documentIds[mdlDoc]
        for mdlDoc in self.modelXbrl.urlDocs.values():
            if mdlDoc.type in (Type.INSTANCE, Type.INLINEXBRL):
                referencedDocuments.add( (self.reportId, self.documentIds[mdlDoc] ))
            if mdlDoc in self.documentIds:
                for refDoc, ref in mdlDoc.referencesDocument.items():
                    if refDoc.inDTS and ref.referenceTypes & {"href", "import", "include"} \
                       and refDoc in self.documentIds:
                        if refDoc.type == Type.SCHEMA:
                            nsAuthority = authority(refDoc.targetNamespace, includeScheme=False)
                            nsPath = refDoc.targetNamespace.split('/')
                            if len(nsPath) > 2:
                                if ((nsAuthority in ("fasb.org", "xbrl.us") and nsPath[-2] == "us-gaap") or
                                    (nsAuthority == "xbrl.ifrs.org" and nsPath[-1] in ("ifrs", "ifrs-full", "ifrs-smes"))):
                                    stdSchemaDocId = self.documentIds[refDoc]
                                elif (nsAuthority == "xbrl.sec.gov" and nsPath[-2] == "rr"):
                                    agencySchemaDocId = self.documentIds[refDoc]


        self.updateTable("report",
                         ("report_id", "report_data_doc_id", "report_schema_doc_id", "agency_schema_doc_id", "standard_schema_doc_id"),
                         ((self.reportId, instDocId, instSchemaDocId, agencySchemaDocId, stdSchemaDocId),)
                        )


    def insertAspects(self):
        self.showStatus("insert aspects")

        # determine new filing documents and types they use
        filingDocumentAspects = set()
        existingDocumentUsedAspects = set()
        for concept in self.modelXbrl.qnameConcepts.values():
            if concept.modelDocument not in self.existingDocumentIds:
                filingDocumentAspects.add(concept)
                filingDocumentAspectType = concept.type
                if filingDocumentAspectType is not None and filingDocumentAspectType not in self.typesUsed:
                        self.typesUsed.add(filingDocumentAspectType)
            elif concept in self.aspectsUsed:
                existingDocumentUsedAspects.add(concept)

        filingDocumentTypes = set()
        existingDocumentUsedTypes = set()
        for modelType in self.modelXbrl.qnameTypes.values():
            if modelType.modelDocument not in self.existingDocumentIds:
                filingDocumentTypes.add(modelType)
            elif modelType in self.typesUsed:
                existingDocumentUsedTypes.add(modelType)


        # get existing element IDs
        self.typeQnameId = {}
        if existingDocumentUsedTypes:
            typeQnameIds = []
            table = self.getTable('data_type', 'data_type_id',
                                  ('document_id', 'qname',),
                                  ('document_id', 'qname',),
                                  tuple((self.documentIds[modelType.modelDocument],
                                         modelType.qname.clarkNotation)
                                        for modelType in existingDocumentUsedTypes
                                        if modelType.modelDocument in self.documentIds),
                                  checkIfExisting=True,
                                  insertIfNotMatched=False)
            for typeId, docId, qn in table:
                self.typeQnameId[qname(qn)] = typeId

        table = self.getTable('data_type', 'data_type_id',
                              ('document_id', 'xml_id', 'xml_child_seq',
                               'qname', 'name', 'base_type', 'derived_from_type_id'),
                              ('document_id', 'qname',),
                              tuple((self.documentIds[modelType.modelDocument],
                                     modelType.id,
                                     elementChildSequence(modelType),
                                     modelType.qname.clarkNotation,
                                     modelType.name,
                                     modelType.baseXsdType,
                                     self.typeQnameId.get(modelType.typeDerivedFrom)
                                     if isinstance(modelType.typeDerivedFrom, ModelType) else None)
                                    for modelType in filingDocumentTypes
                                    if modelType.modelDocument in self.documentIds)
                             )
        for typeId, docId, qn in table:
            self.typeQnameId[qname(qn)] = typeId

        updatesToDerivedFrom = set()
        for modelType in filingDocumentTypes:
            if isinstance(modelType.typeDerivedFrom, ModelType):
                typeDerivedFrom = modelType.typeDerivedFrom
                if (typeDerivedFrom in filingDocumentTypes and
                    modelType.qname in self.typeQnameId and
                    typeDerivedFrom.qname in self.typeQnameId):
                    updatesToDerivedFrom.add( (self.typeQnameId[modelType.qname],
                                               self.typeQnameId[typeDerivedFrom.qname]) )
        # update derivedFrom's of newly added types
        if updatesToDerivedFrom:
            self.updateTable('data_type',
                             ('data_type_id', 'derived_from_type_id'),
                             updatesToDerivedFrom)

        existingDocumentUsedTypes.clear() # dereference
        filingDocumentTypes.clear() # dereference

        self.aspectQnameId = {}

        # get existing element IDs
        if existingDocumentUsedAspects:
            table = self.getTable('aspect', 'aspect_id',
                                  ('document_id', 'qname',),
                                  ('document_id', 'qname',),
                                  tuple((self.documentIds[concept.modelDocument],
                                         concept.qname.clarkNotation)
                                        for concept in existingDocumentUsedAspects
                                        if concept.modelDocument in self.documentIds),
                                  checkIfExisting=True,
                                  insertIfNotMatched=False)
            for aspectId, docId, qn in table:
                self.aspectQnameId[qname(qn)] = aspectId

        aspects = []
        for concept in filingDocumentAspects:
            niceType  = concept.niceType
            if niceType is not None and len(niceType) > 128:
                niceType = niceType[:128]
            if concept.modelDocument in self.documentIds:
                aspects.append((self.documentIds[concept.modelDocument],
                                 concept.id,
                                 elementChildSequence(concept),
                                 concept.qname.clarkNotation,
                                 concept.name,
                                 self.typeQnameId.get(concept.typeQname),
                                 niceType[:128] if niceType is not None else None,
                                 self.aspectQnameId.get(concept.substitutionGroupQname),
                                 concept.balance,
                                 concept.periodType,
                                 concept.isAbstract,
                                 concept.isNillable,
                                 concept.isNumeric,
                                 concept.isMonetary,
                                 concept.isTextBlock))
        table = self.getTable('aspect', 'aspect_id',
                              ('document_id', 'xml_id', 'xml_child_seq',
                               'qname', 'name', 'datatype_id', 'base_type', 'substitution_group_aspect_id',
                               'balance', 'period_type', 'abstract', 'nillable',
                               'is_numeric', 'is_monetary', 'is_text_block'),
                              ('document_id', 'qname'),
                              aspects
                             )
        for aspectId, docId, qn in table:
            self.aspectQnameId[qname(qn)] = aspectId

        updatesToSubstitutionGroup = set()
        for concept in filingDocumentAspects:
            if concept.substitutionGroup in filingDocumentAspects and concept.modelDocument in self.documentIds:
                updatesToSubstitutionGroup.add( (self.aspectQnameId[concept.qname],
                                                 self.aspectQnameId.get(concept.substitutionGroupQname)) )
        # update derivedFrom's of newly added types
        if updatesToSubstitutionGroup:
            self.updateTable('aspect',
                             ('aspect_id', 'substitution_group_aspect_id'),
                             updatesToSubstitutionGroup)

        filingDocumentAspects.clear() # dereference
        existingDocumentUsedAspects.clear() # dereference

    def insertArcroleTypes(self):
        self.showStatus("insert arcrole types")
        # add existing arcrole types
        arcroleTypesByIds = set((self.documentIds[arcroleType.modelDocument],
                                 arcroleType.roleURI) # key on docId, uriId
                                for arcroleTypes in self.modelXbrl.arcroleTypes.values()
                                for arcroleType in arcroleTypes
                                if arcroleType.modelDocument in self.existingDocumentIds)
        table = self.getTable('arcrole_type', 'arcrole_type_id',
                              ('document_id', 'arcrole_uri'),
                              ('document_id', 'arcrole_uri'),
                              tuple((arcroleTypeIDs[0], # doc Id
                                     arcroleTypeIDs[1] # uri Id
                                     )
                                    for arcroleTypeIDs in arcroleTypesByIds),
                              checkIfExisting=True,
                              insertIfNotMatched=False)
        self.arcroleTypeIds = {}
        for arcroleId, docId, uri in table:
            self.arcroleTypeIds[(docId, uri)] = arcroleId

        # added document arcrole type
        arcroleTypesByIds = dict(((self.documentIds[arcroleType.modelDocument],
                                   arcroleType.arcroleURI), # key on docId, uriId
                                  arcroleType) # value is roleType object
                                 for arcroleTypes in self.modelXbrl.arcroleTypes.values()
                                 for arcroleType in arcroleTypes
                                 if arcroleType.modelDocument not in self.existingDocumentIds)
        table = self.getTable('arcrole_type', 'arcrole_type_id',
                              ('document_id', 'xml_id', 'xml_child_seq', 'arcrole_uri', 'cycles_allowed', 'definition'),
                              ('document_id', 'arcrole_uri'),
                              tuple((arcroleTypeIDs[0], # doc Id
                                     arcroleType.id,
                                     elementChildSequence(arcroleType),
                                     arcroleType.arcroleURI,
                                     arcroleType.cyclesAllowed,
                                     arcroleType.definition)
                                    for arcroleTypeIDs, arcroleType in arcroleTypesByIds.items()))

        for arcroleId, docId, uri in table:
            self.arcroleTypeIds[(docId, uri)] = arcroleId

        table = self.getTable('used_on',
                              None, # no record id in this table
                              ('object_id', 'aspect_id'),
                              ('object_id', 'aspect_id'),
                              tuple((self.arcroleTypeIds[(arcroleTypeIDs[0], arcroleType.arcroleURI)],
                                     self.aspectQnameId[usedOnQn])
                                    for arcroleTypeIDs, arcroleType in arcroleTypesByIds.items()
                                    for usedOnQn in arcroleType.usedOns
                                    if usedOnQn in self.aspectQnameId),
                              checkIfExisting=True)

    def insertRoleTypes(self):
        self.showStatus("insert role types")
        # add existing role types
        roleTypesByIds = set((self.documentIds[roleType.modelDocument],
                              roleType.roleURI) # key on docId, uriId
                              for roleTypes in self.modelXbrl.roleTypes.values()
                              for roleType in roleTypes
                              if roleType.modelDocument in self.existingDocumentIds)
        table = self.getTable('role_type', 'role_type_id',
                              ('document_id', 'role_uri'),
                              ('document_id', 'role_uri'),
                              tuple((roleTypeIDs[0], # doc Id
                                     roleTypeIDs[1] # uri Id
                                     )
                                    for roleTypeIDs in roleTypesByIds),
                              checkIfExisting=True,
                              insertIfNotMatched=False)
        self.roleTypeIds = {}
        for roleId, docId, uri in table:
            self.roleTypeIds[(docId, uri)] = roleId

        # new document role types
        roleTypesByIds = dict(((self.documentIds[roleType.modelDocument],
                                roleType.roleURI), # key on docId, uriId
                               roleType) # value is roleType object
                              for roleTypes in self.modelXbrl.roleTypes.values()
                              for roleType in roleTypes
                              if roleType.modelDocument not in self.existingDocumentIds)
        table = self.getTable('role_type', 'role_type_id',
                              ('document_id', 'xml_id', 'xml_child_seq', 'role_uri', 'definition'),
                              ('document_id', 'role_uri'),
                              tuple((roleTypeIDs[0], # doc Id
                                     roleType.id,
                                     elementChildSequence(roleType),
                                     roleTypeIDs[1], # uri Id
                                     roleType.definition)
                                    for roleTypeIDs, roleType in roleTypesByIds.items()))
        for roleId, docId, uri in table:
            self.roleTypeIds[(docId, uri)] = roleId


        table = self.getTable('used_on',
                              None, # no record id in this table
                              ('object_id', 'aspect_id'),
                              ('object_id', 'aspect_id'),
                              tuple((self.roleTypeIds[(roleTypeIDs[0], roleType.roleURI)],
                                     self.aspectQnameId[usedOnQn])
                                    for roleTypeIDs, roleType in roleTypesByIds.items()
                                    for usedOnQn in roleType.usedOns
                                    if usedOnQn in self.aspectQnameId),
                              checkIfExisting=True)

    def insertResources(self):
        self.showStatus("insert resources")
        # deduplicate resources (may be on multiple arcs)
        arcroles = [arcrole
                    # check whether relationship_set is completely in instance or part/all in taxonomy
                    for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys()
                    if ELR is None and linkqname is None and arcqname is None and not arcrole.startswith("XBRL-")
                       and self.arcroleHasResource[arcrole]
                       and (self.arcroleInInstance[arcrole] or not self.isExistingTaxonomyRelSetsOwner)]
        # note that lxml has no column numbers, use objectIndex as pseudo-column number
        uniqueResources = dict(((self.documentIds[resource.modelDocument],
                                 resource.objectIndex), resource)
                               for arcrole in arcroles
                               for rel in self.modelXbrl.relationshipSet(arcrole).modelRelationships
                               for resource in (rel.fromModelObject, rel.toModelObject)
                               if isinstance(resource, ModelResource))
        table = self.getTable('resource', 'resource_id',
                              ('document_id', 'xml_id', 'xml_child_seq', 'qname', 'role', 'value', 'xml_lang'),
                              ('document_id', 'xml_child_seq'),
                              tuple((self.documentIds[resource.modelDocument],
                                     resource.id,
                                     elementChildSequence(resource),
                                     resource.qname.clarkNotation,
                                     resource.role,
                                     resource.stringValue,
                                     resource.xmlLang)
                                    for resource in uniqueResources.values()),
                              checkIfExisting=True)
        self.resourceId = dict(((docId, xml_child_seq), id)
                               for id, docId, xml_child_seq in table)
        uniqueResources.clear()


    def modelObjectId(self, modelObject):
        if isinstance(modelObject, ModelConcept):
            return self.aspectQnameId.get(modelObject.qname)
        elif isinstance(modelObject, ModelType):
            return self.aspectTypeIds.get(modelObject.qname)
        elif isinstance(modelObject, ModelResource):
            return self.resourceId.get((self.documentIds[modelObject.modelDocument],
                                        elementChildSequence(modelObject)))
        elif isinstance(modelObject, ModelFact):
            return self.factDataPointId.get((self.documentIds[modelObject.modelDocument],
                                             elementChildSequence(modelObject)))
        else:
            return None

    def insertRelationships(self):
        self.showStatus("insert relationship sets")
        table = self.getTable('relationship_set', 'relationship_set_id',
                              ('document_id', 'link_role', 'arc_role', 'link_qname', 'arc_qname'),
                              ('document_id', 'link_role', 'arc_role', 'link_qname', 'arc_qname'),
                              tuple((self.documentIds[self.modelXbrl.modelDocument if self.arcroleInInstance[arcrole]
                                                      else self.taxonomyRelSetsOwner],
                                     ELR,
                                     arcrole,
                                     linkqname.clarkNotation,
                                     arcqname.clarkNotation)
                                    for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys()
                                    if ELR and linkqname and arcqname and not arcrole.startswith("XBRL-")
                                           and (not self.isExistingTaxonomyRelSetsOwner or self.arcroleInInstance[arcrole])))
        self.relSetId = dict(((linkRole, arcRole, lnkQn, arcQn), id)
                             for id, document_id, linkRole, arcRole, lnkQn, arcQn in table)
        # do tree walk to build relationships with depth annotated, no targetRole navigation
        dbRels = []

        def walkTree(rels, seq, depth, relationshipSet, visited, dbRels, relSetId):
            for rel in rels:
                if rel not in visited and isinstance(rel.toModelObject, ModelObject):
                    visited.add(rel)
                    dbRels.append((rel, seq, depth, relSetId))
                    seq += 1
                    seq = walkTree(relationshipSet.fromModelObject(rel.toModelObject), seq, depth+1, relationshipSet, visited, dbRels, relSetId)
                    visited.remove(rel)
            return seq

        for arcrole, ELR, linkqname, arcqname in self.modelXbrl.baseSets.keys():
            if (ELR and linkqname and arcqname and not arcrole.startswith("XBRL-")
                and (not self.isExistingTaxonomyRelSetsOwner or self.arcroleInInstance[arcrole])):
                relSetId = self.relSetId[(ELR,
                                          arcrole,
                                          linkqname.clarkNotation,
                                          arcqname.clarkNotation)]
                relationshipSet = self.modelXbrl.relationshipSet(arcrole, ELR, linkqname, arcqname)
                seq = 1
                for rootConcept in relationshipSet.rootConcepts:
                    seq = walkTree(relationshipSet.fromModelObject(rootConcept), seq, 1, relationshipSet, set(), dbRels, relSetId)

        def resourceResourceId(resource):
            if isinstance(resource, ModelResource):
                return self.resourceId.get((self.documentIds[resource.modelDocument],
                                            resource.sourceline,
                                            resource.objectIndex))
            else:
                return None

        table = self.getTable('relationship', 'relationship_id',
                              ('document_id', 'xml_id', 'xml_child_seq',
                               'relationship_set_id', 'reln_order',
                               'from_id', 'to_id', 'calculation_weight',
                               'tree_sequence', 'tree_depth', 'preferred_label_role'),
                              ('relationship_set_id', 'document_id', 'xml_child_seq'),
                              tuple((self.documentIds[rel.modelDocument],
                                     rel.id,
                                     elementChildSequence(rel.arcElement),
                                     relSetId,
                                     self.dbNum(rel.order),
                                     self.modelObjectId(rel.fromModelObject),
                                     self.modelObjectId(rel.toModelObject),
                                     self.dbNum(rel.weight), # none if no weight
                                     sequence,
                                     depth,
                                     rel.preferredLabel)
                                    for rel, sequence, depth, relSetId in dbRels
                                    if isinstance(rel.fromModelObject, ModelObject) and isinstance(rel.toModelObject, ModelObject)))
        self.relationshipId = dict(((docId,xml_child_seq), relationshipId)
                                   for relationshipId, relSetId, docId, xml_child_seq in table)
        table = self.getTable('root', None,
                              ('relationship_set_id', 'relationship_id'),
                              ('relationship_set_id', 'relationship_id'),
                              tuple((relSetId,
                                     self.relationshipId[self.documentIds[rel.modelDocument],
                                                         elementChildSequence(rel.arcElement)])
                                    for rel, sequence, depth, relSetId in dbRels
                                    if depth == 1 and
                                       isinstance(rel.fromModelObject, ModelObject) and isinstance(rel.toModelObject, ModelObject)))
        del dbRels[:]   # dererefence

    def insertDataPoints(self):
        reportId = self.reportId
        if self.filingPreviouslyInDB:
            self.showStatus("deleting prior data points of this report")
            # remove prior facts
            self.lockTables(("data_point", "entity_identifier", "period", "aspect_value_selection",
                             "aspect_value_selection_set", "unit_measure", "unit",
                             "table_data_points"))
            self.execute("DELETE FROM {0} WHERE {0}.report_id = {1}"
                         .format( self.dbTableName("data_point"), reportId),
                         close=False, fetch=False)
            self.execute("DELETE FROM {0} WHERE {0}.report_id = {1}"
                         .format( self.dbTableName("entity_identifier"), reportId),
                         close=False, fetch=False)
            self.execute("DELETE FROM {0} WHERE {0}.report_id = {1}"
                         .format( self.dbTableName("period"), reportId),
                         close=False, fetch=False)
            self.execute("DELETE from {0} "
                         "USING {1} "
                         "WHERE {1}.report_id = {2} AND {0}.aspect_value_selection_id = {1}.aspect_value_selection_id"
                         .format( self.dbTableName("aspect_value_selection"),
                                  self.dbTableName("aspect_value_selection_set"),
                                  reportId),
                         close=False, fetch=False)
            self.execute("DELETE FROM {0} WHERE {0}.report_id = {1};"
                         .format( self.dbTableName("aspect_value_selection_set"), reportId),
                         close=False, fetch=False)
            self.execute("DELETE from {0} "
                         "USING {1} "
                         "WHERE {1}.report_id = {2} AND {0}.unit_id = {1}.unit_id"
                         .format( self.dbTableName("unit_measure"),
                                  self.dbTableName("unit"),
                                  reportId),
                         close=False, fetch=False)
            self.execute("DELETE from {0} WHERE {0}.report_id = {1}"
                         .format( self.dbTableName("unit"), reportId),
                         close=False, fetch=False)
            self.execute("DELETE FROM {0} WHERE {0}.report_id = {1}"
                         .format( self.dbTableName("table_data_points"), reportId),
                         close=False, fetch=False)
        self.showStatus("insert data points")
        # units
        table = self.getTable('unit', 'unit_id',
                              ('report_id', 'xml_id', 'xml_child_seq', 'measures_hash'),
                              ('report_id', 'measures_hash'),
                              tuple((reportId,
                                     unit.id,
                                     elementChildSequence(unit),
                                     unit.md5hash)
                                    for unit in dict((unit.md5hash,unit) # deduplicate by md5hash
                                                     for unit in self.modelXbrl.units.values()).values()))
        self.unitId = dict(((_reportId, measuresHash), id)
                           for id, _reportId, measuresHash in table)
        # measures
        table = self.getTable('unit_measure',
                              None,
                              ('unit_id', 'qname', 'is_multiplicand'),
                              ('unit_id', 'qname', 'is_multiplicand'),
                              tuple((self.unitId[(reportId,unit.md5hash)],
                                     measure.clarkNotation,
                                     i == 0)
                                    for unit in self.modelXbrl.units.values()
                                    for i in range(2)
                                    for measure in unit.measures[i]))
        table = self.getTable('entity_identifier', 'entity_identifier_id',
                              ('report_id', 'scheme', 'identifier'),
                              ('report_id', 'scheme', 'identifier'),
                              set((reportId,
                                   cntx.entityIdentifier[0],
                                   cntx.entityIdentifier[1])
                                for cntx in self.modelXbrl.contexts.values()),
                              checkIfExisting=True) # entities shared across multiple instance/inline docs
        self.entityIdentifierId = dict(((_reportId, entScheme, entIdent), id)
                                       for id, _reportId, entScheme, entIdent in table)
        table = self.getTable('period', 'period_id',
                              ('report_id', 'start_date', 'end_date', 'is_instant', 'is_forever'),
                              ('report_id', 'start_date', 'end_date', 'is_instant', 'is_forever'),
                              set((reportId,
                                   cntx.startDatetime if cntx.isStartEndPeriod else None,
                                   cntx.endDatetime if (cntx.isStartEndPeriod or cntx.isInstantPeriod) else None,
                                   cntx.isInstantPeriod,
                                   cntx.isForeverPeriod)
                                for cntx in self.modelXbrl.contexts.values()),
                              checkIfExisting=True) # periods shared across multiple instance/inline docs
        self.periodId = dict(((_reportId, start, end, isInstant, isForever), id)
                             for id, _reportId, start, end, isInstant, isForever in table)

        def cntxDimsSet(cntx):
            return frozenset((self.aspectQnameId[modelDimValue.dimensionQname],
                              self.aspectQnameId.get(modelDimValue.memberQname),
                              modelDimValue.isTyped,
                              modelDimValue.stringValue if modelDimValue.isTyped else None)
                             for modelDimValue in cntx.qnameDims.values()
                             if modelDimValue.dimensionQname in self.aspectQnameId)

        cntxAspectValueSelectionSet = dict((cntx, cntxDimsSet(cntx))
                                            for cntx in self.modelXbrl.contexts.values())

        aspectValueSelections = set(aspectValueSelectionSet
                                    for cntx, aspectValueSelectionSet in cntxAspectValueSelectionSet.items()
                                    if aspectValueSelectionSet)
        self.lockTables(("aspect_value_selection_set",))
        self.execute("DELETE FROM {0} WHERE report_id = {1}"
                     .format(self.dbTableName("aspect_value_selection_set"), reportId),
                     close=False, fetch=False)
        table = self.getTable('aspect_value_selection_set', 'aspect_value_selection_id',
                              ('report_id', ),
                              ('report_id', ),
                              tuple((reportId,)
                                    for aspectValueSelection in aspectValueSelections)
                              )
        # assure we only get single entry per result (above gives cross product)
        table = self.execute("SELECT aspect_value_selection_id, report_id from {0} "
                             "WHERE report_id = {1}"
                             .format(self.dbTableName("aspect_value_selection_set"), reportId))
        aspectValueSelectionSets = dict((aspectValueSelections.pop(), id)
                                        for id, _reportId in table)

        cntxAspectValueSelectionSetId = dict((cntx, aspectValueSelectionSets[_cntxDimsSet])
                                             for cntx, _cntxDimsSet in cntxAspectValueSelectionSet.items()
                                             if _cntxDimsSet)

        table = self.getTable('aspect_value_selection',
                              None,
                              ('aspect_value_selection_id', 'aspect_id', 'aspect_value_id', 'is_typed_value', 'typed_value'),
                              ('aspect_value_selection_id', ),
                              tuple((aspectValueSetId, dimId, dimMbrId, isTyped, typedValue)
                                    for aspectValueSelection, aspectValueSetId in aspectValueSelectionSets.items()
                                    for dimId, dimMbrId, isTyped, typedValue in aspectValueSelection)
                              )

        # facts
        def insertFactSet(modelFacts, parentDatapointId):
            facts = []
            for fact in modelFacts:
                if fact.concept is not None and getattr(fact, "xValid", UNVALIDATED) >= VALID and fact.qname is not None:
                    cntx = fact.context
                    documentId = self.documentIds[fact.modelDocument]
                    facts.append((reportId,
                                  documentId,
                                  fact.id,
                                  elementChildSequence(fact),
                                  fact.sourceline,
                                  parentDatapointId, # parent ID
                                  self.aspectQnameId.get(fact.qname),
                                  fact.contextID,
                                  self.entityIdentifierId.get((reportId, cntx.entityIdentifier[0], cntx.entityIdentifier[1]))
                                      if cntx is not None else None,
                                      self.periodId.get((reportId,
                                                    cntx.startDatetime if cntx.isStartEndPeriod else None,
                                                    cntx.endDatetime if (cntx.isStartEndPeriod or cntx.isInstantPeriod) else None,
                                                    cntx.isInstantPeriod,
                                                    cntx.isForeverPeriod)) if cntx is not None else None,
                                  cntxAspectValueSelectionSetId.get(cntx) if cntx is not None else None,
                                  self.unitId.get((reportId,fact.unit.md5hash)) if fact.unit is not None else None,
                                  fact.isNil,
                                  fact.precision,
                                  fact.decimals,
                                  roundValue(fact.value, fact.precision, fact.decimals) if fact.isNumeric and not fact.isNil else None,
                                  fact.value
                                  ))
            table = self.getTable('data_point', 'datapoint_id',
                                  ('report_id', 'document_id', 'xml_id', 'xml_child_seq', 'source_line',
                                   'parent_datapoint_id',  # tuple
                                   'aspect_id',
                                   'context_xml_id', 'entity_identifier_id', 'period_id', 'aspect_value_selection_id', 'unit_id',
                                   'is_nil', 'precision_value', 'decimals_value', 'effective_value', 'value'),
                                  ('document_id', 'xml_child_seq'),
                                  facts)
            xmlIdDataPointId = dict(((docId, xml_child_seq), datapointId)
                                    for datapointId, docId, xml_child_seq in table)
            self.factDataPointId.update(xmlIdDataPointId)
            for fact in modelFacts:
                if fact.isTuple:
                    try:
                        insertFactSet(fact.modelTupleFacts,
                                      xmlIdDataPointId[(self.documentIds[fact.modelDocument],
                                                        elementChildSequence(fact))])
                    except KeyError:
                        self.modelXbrl.info("xpDB:warning",
                                            _("Loading XBRL DB: tuple's datapoint not found: %(tuple)s"),
                                            modelObject=fact, tuple=fact.qname)

        self.factDataPointId = {}
        insertFactSet(self.modelXbrl.facts, None)
        # hashes
        if self.tableFacts: # if any entries
            tableDataPoints = []
            for roleType, tableCode, fact in self.tableFacts:
                try:
                    tableDataPoints.append((reportId,
                                            self.roleTypeIds[(self.documentIds[roleType.modelDocument],
                                                              roleType.roleURI)],
                                            tableCode,
                                            self.factDataPointId[(self.documentIds[fact.modelDocument],
                                                                  elementChildSequence(fact))]))
                except KeyError:
                    # print ("missing table data points role or data point")
                    pass
            table = self.getTable('table_data_points', None,
                                  ('report_id', 'object_id', 'table_code', 'datapoint_id'),
                                  ('report_id', 'object_id', 'datapoint_id'),
                                  tableDataPoints)

    def insertValidationResults(self):
        reportId = self.reportId

        if self.filingPreviouslyInDB:
            self.showStatus("deleting prior messages of this report")
            # remove prior messages for this report
            self.lockTables(("message", "message_reference"))
            self.execute("DELETE from {0} "
                         "USING {1} "
                         "WHERE {1}.report_id = {2} AND {1}.message_id = {0}.message_id"
                         .format(self.dbTableName("message_reference"),
                                 self.dbTableName("message"),
                                 reportId),
                         close=False, fetch=False)
            self.execute("DELETE FROM {} WHERE message.report_id = {}"
                         .format(self.dbTableName("message"),reportId),
                         close=False, fetch=False)
        messages = []
        messageRefs = defaultdict(set) # direct link to objects
        for i, logEntry in enumerate(self.loggingEntries):
            sequenceInReport = i+1
            for ref in logEntry['refs']:
                modelObject = self.modelXbrl.modelObject(ref.get('objectId',''))
                # for now just find a concept
                objectId = None
                if isinstance(modelObject, ModelFact):
                    objectId = self.factDataPointId.get((self.documentIds.get(modelObject.modelDocument),
                                                         elementChildSequence(modelObject)))
                elif isinstance(modelObject, ModelRelationship):
                    objectId = self.relSetId.get((modelObject.linkrole,
                                                  modelObject.arcrole,
                                                  modelObject.linkQname.clarkNotation,
                                                  modelObject.arcElement.qname.clarkNotation))
                elif isinstance(modelObject, ModelConcept):
                    objectId = self.aspectQnameId.get(modelObject.qname)
                elif isinstance(modelObject, ModelXbrl):
                    objectId = reportId
                elif hasattr(modelObject, "modelDocument"):
                    objectId = self.documentIds.get(modelObject.modelDocument)

                if objectId is not None:
                    messageRefs[sequenceInReport].add(objectId)

            messages.append((reportId,
                             sequenceInReport,
                             logEntry['code'],
                             logEntry['level'],
                             logEntry['message']['text']))
        if messages:
            self.showStatus("insert validation messages")
            table = self.getTable('message', 'message_id',
                                  ('report_id', 'sequence_in_report', 'message_code', 'message_level', 'value'),
                                  ('report_id', 'sequence_in_report'),
                                  messages)
            messageIds = dict((sequenceInReport, messageId)
                              for messageId, _reportId, sequenceInReport in table)
            table = self.getTable('message_reference', None,
                                  ('message_id', 'object_id'),
                                  ('message_id', 'object_id'),
                                  tuple((messageId,
                                         objectId)
                                        for sequenceInReport, objectIds in messageRefs.items()
                                        for objectId in objectIds
                                        for messageId in (messageIds[sequenceInReport],)))

countryOfState = {
    "AL": "US","AK": "US","AZ": "US","AR": "US","CA": "US","CO": "US", "CT": "US","DE": "US",
    "FL": "US","GA": "US","HI": "US","ID": "US","IL": "US","IN": "US","IA": "US","KS": "US",
    "KY": "US","LA": "US","ME": "US","MD": "US","MA": "US","MI": "US","MN": "US","MS": "US",
    "MO": "US","MT": "US","NE": "US","NV": "US","NH": "US","NJ": "US","NM": "US","NY": "US",
    "NC": "US","ND": "US","OH": "US","OK": "US","OR": "US","PA": "US","RI": "US","SC": "US",
    "SD": "US","TN": "US","TX": "US","UT": "US","VT": "US","VA": "US","WA": "US","WV": "US",
    "WI": "US","WY": "US","DC": "US","PR": "US","VI": "US","AS": "US","GU": "US","MP": "US",
    "AB": "CA","BC": "CA","MB": "CA","NB": "CA","NL": "CA","NS": "CA","ON": "CA","PE": "CA",
    "QC": "CA","SK": "CA","NT": "CA","NU": "CA","YT": "CA"}



"""
See COPYRIGHT.md for copyright information.
"""

from typing import TYPE_CHECKING

from arelle.PythonUtil import OrderedSet
from .XbrlProperty import XbrlProperty
from .XbrlTypes import XbrlTaxonomyType, SQNameKeyType
from .ModelValueMore import SQName
from .XbrlTaxonomyObject import XbrlReferencableTaxonomyObject

class XbrlEntity(XbrlReferencableTaxonomyObject):
    taxonomy: XbrlTaxonomyType
    name: SQNameKeyType # (required) The entity SQName that identifies the entity so it can be referenced by other objects.
    properties: OrderedSet[XbrlProperty] # (optional) ordered set of property objects used to specify additional properties associated with the concept using the property object. Only immutable properties as defined in the propertyType object can be added to a concept.



'''
See COPYRIGHT.md for copyright information.
'''
from tkinter import Toplevel, N, S, E, W, PhotoImage
try:
    from tkinter.ttk import Frame, Button
except ImportError:
    from ttk import Frame, Button
import os
import regex as re
from arelle.UiUtil import gridHdr, gridCell, gridCombobox, label, checkbox
from arelle.CntlrWinTooltip import ToolTip
from arelle import XbrlConst

'''
caller checks accepted, if True, caller retrieves url
'''
def getArcroleGroup(mainWin, modelXbrl):
    dialog = DialogArcroleGroup(mainWin, modelXbrl)
    return dialog.selectedGroup


class DialogArcroleGroup(Toplevel):
    def __init__(self, mainWin, modelXbrl):
        parent = mainWin.parent
        super(DialogArcroleGroup, self).__init__(parent)
        self.mainWin = mainWin
        self.parent = parent
        self.modelXbrl = modelXbrl
        parentGeometry = re.match(r"(\d+)x(\d+)[+]?([-]?\d+)[+]?([-]?\d+)", parent.geometry())
        dialogX = int(parentGeometry.group(3))
        dialogY = int(parentGeometry.group(4))
        self.selectedGroup = None

        self.transient(self.parent)
        self.title(_("Select Arcrole Group"))

        frame = Frame(self)

        '''
        dialogFrame = Frame(frame, width=500)
        dialogFrame.columnconfigure(0, weight=1)
        dialogFrame.rowconfigure(0, weight=1)
        dialogFrame.grid(row=0, column=0, columnspan=4, sticky=(N, S, E, W), padx=3, pady=3)
        '''

        # mainWin.showStatus(_("loading formula options and parameters"))

        # load grid
        groupLabel = label(frame, 1, 0, _("Group:"))
        self.arcroleGroups = mainWin.config.get("arcroleGroups", {})
        arcroleGroupSelected = self.mainWin.config.get("arcroleGroupSelected")
        if arcroleGroupSelected in self.arcroleGroups:
            arcroleGroup = self.arcroleGroups[arcroleGroupSelected]
        else:
            arcroleGroup = []
            arcroleGroupSelected = None
        self.groupName = gridCombobox(frame, 2, 0,
                                      value=arcroleGroupSelected,
                                      values=sorted(self.arcroleGroups.keys()),
                                      comboboxselected=self.comboBoxSelected)
        groupToolTipMessage = _("Select an existing arcrole group, or enter a name for a new arcrole group.  "
                                "If selecting an existing group, it can be edited, and changes will be saved in the config file.  "
                                "If nothing is changed for an existing group, the saved setting is not disturbed.  "
                                "Arcroles with checkboxes below are shown only for arcroles that have relationships in the loaded DTS, "
                                "but if an existing group is selected with more arcroles (that were not in the current DTS) then "
                                "the prior setting with not-present arcroles is preserved. ")
        ToolTip(self.groupName, text=groupToolTipMessage, wraplength=360)
        ToolTip(groupLabel, text=groupToolTipMessage, wraplength=360)
        clearImage = PhotoImage(file=os.path.join(mainWin.imagesDir, "toolbarDelete.gif"))
        clearGroupNameButton = Button(frame, image=clearImage, width=12, command=self.clearGroupName)
        clearGroupNameButton.grid(row=0, column=3, sticky=W)
        ToolTip(clearGroupNameButton, text=_("Remove the currently selected arcrole group from the config file. "
                                             "After removing, you may select another arcrole, but must select 'OK' for the "
                                             "removal to be saved. "),
                wraplength=240)
        arcrolesLabel = label(frame, 1, 1, _("Arcroles:"))
        ToolTip(arcrolesLabel, text=_("Shows all the arcroles that are present in this DTS. "),
                wraplength=240)
        from arelle.ModelRelationshipSet import baseSetArcroles
        self.options = {}
        self.checkboxes = []
        y = 1
        for name, arcrole in baseSetArcroles(self.modelXbrl):
            if arcrole.startswith("http://"):
                self.options[arcrole] = arcrole in arcroleGroup
                self.checkboxes.append(
                   checkbox(frame, 2, y,
                            name[1:],
                            arcrole,
                            columnspan=2)
                )
                y += 1

        mainWin.showStatus(None)

        self.options[XbrlConst.arcroleGroupDetect] = XbrlConst.arcroleGroupDetect in arcroleGroup
        self.autoOpen = checkbox(frame, 1, y, _("detect"), XbrlConst.arcroleGroupDetect)
        self.autoOpen.grid(sticky=W, columnspan=2)
        self.checkboxes.append(self.autoOpen)
        ToolTip(self.autoOpen, text=_("If checked, this arcrole group will be detected if any arcrole of the group is present in a DTS, for example to open a treeview pane. "),
                wraplength=240)
        okButton = Button(frame, text=_("OK"), width=8, command=self.ok)
        cancelButton = Button(frame, text=_("Cancel"), width=8, command=self.close)
        cancelButton.grid(row=y, column=1, sticky=E, columnspan=3, pady=3, padx=3)
        okButton.grid(row=y, column=1, sticky=E, columnspan=3, pady=3, padx=64)
        ToolTip(okButton, text=_("Open a treeview with named arcrole group and selected arcroles. "
                                 "If any changes were made to checkboxes or name, save in the config. "),
                wraplength=240)
        ToolTip(cancelButton, text=_("Close this dialog, without saving arcrole group changes or opening a view pane. "),
                wraplength=240)

        frame.grid(row=0, column=0, sticky=(N,S,E,W))
        frame.columnconfigure(1, weight=3)
        frame.columnconfigure(2, weight=1)
        frame.columnconfigure(3, weight=3)
        window = self.winfo_toplevel()
        window.columnconfigure(0, weight=1)
        self.geometry("+{0}+{1}".format(dialogX+50,dialogY+100))

        #self.bind("<Return>", self.ok)
        #self.bind("<Escape>", self.close)

        self.protocol("WM_DELETE_WINDOW", self.close)
        self.grab_set()
        self.wait_window(self)

    def ok(self, event=None):
        groupName = self.groupName.value
        arcrolesSelected = [checkbox.attr for checkbox in self.checkboxes if checkbox.value]
        if groupName:
            self.mainWin.config["arcroleGroupSelected"] = groupName
            if groupName not in self.arcroleGroups or any(checkbox.isChanged for checkbox in self.checkboxes):
                self.arcroleGroups[groupName] = arcrolesSelected
                self.mainWin.config["arcroleGroups"] = self.arcroleGroups
                self.mainWin.saveConfig()
            self.selectedGroup = (groupName, arcrolesSelected)
        self.close()

    def close(self, event=None):
        self.parent.focus_set()
        self.destroy()

    def comboBoxSelected(self, *args):
        arcroles = self.arcroleGroups.get(self.groupName.value, [])
        for checkbox in self.checkboxes:
            checkbox.valueVar.set( checkbox.attr in arcroles )
            checkbox.isChanged = False

    def clearGroupName(self):
        groupName = self.groupName.value
        if groupName and groupName in self.arcroleGroups:
            del self.arcroleGroups[groupName]
        self.groupName.valueVar.set('')
        self.groupName["values"] = sorted(self.arcroleGroups.keys())
        for checkbox in self.checkboxes:
            checkbox.valueVar.set( False )
            checkbox.isChanged = False
